#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 26/02/2025 11:03:05

@author: psakic
"""

import os
import numpy as np
from geodezyx import conv, utils
import pandas as pd
import yaml
import copy
import shutil
import datetime as dt

#### Import the logger
import logging
log = logging.getLogger('geodezyx')


#################### DOUBLE DIFF ####################


def gen_dirs_from_double_diff(
    dd_files_paths_in,
    director_generik_path,
    director_name_prefix,
    out_director_folder="",
    temp_data_folder="",
    stations_file="",
    oceanload_file="",
    auto_staocl=False,
    perso_orbclk=False,
    gins_style_orb="GRG",
    ac="igs",
    repro=2,
    auto_interval=True,
    out_coords="NULL",
):
    """
    NEW FCT WHICH CAN MANAGE BOTH ONE RINEX OR A LIST OF RINEX

    produce a director from a rinex, and write it in the specific location

    Input:
        rinex_paths_in : can be a rinex path (a string) or a list of paths

    Return :
            the path of the produced director as a string

    If the rinex is not in a gins style folder, it will be copied in the
    ``temp_data_folder``

    If ``temp_data_folder`` is not specified ( == '') rinex will be copied in
    a ad hoc folder ``../gin/TEMP_DATA``

    If ``out_director_folder`` is not specified ( == ''), output director will be
    created in the ``../gin/data/directeur folder``

    auto_staocl : True or False
        create automatically a station file and a ocean loading file
        with the rinex header
        auto mode is prioritary upon the manu mode
        so, if a path for stat file or ocload file is specified and
        auto_staocl is on, it will be the automatic stat/ocload
        which will be used
    perso_orbclk :  True or False
        download and use specifics orbits
        according to ac & repro args
        (they are useless if perso_orbclk aren't activated)
    auto_interval : True or False
        find the interval in RINEX and apply it to the director
    out_coords : 'XYZ' or 'FLH'/'PLH'
        for geocentrical or geographical coords in output.
        any other string leaves the type of the generic director.
    prairie : True or False
        run prairie externally
        if True, prairie_kwargs control the arguments of the function
        prairie_manual (cf above)
    """

    # Multi or Single Mode ?
    if type(dd_files_paths_in) is list:
        multimode = True
        dd_files_paths_in = dd_files_paths_in
        print("******** DIRECTORS GENRATION ********")
    elif type(dd_files_paths_in) is str:
        multimode = False
        dd_files_paths_in = [dd_files_paths_in]
    else:
        print("ERR : gen_dirs_from_double_diff : check the dd_files_paths_in !!!")
        return None

    N = len(dd_files_paths_in)
    director_output_path_lis = []
    failed_dd_date_lis = []

    for i, dd_path in enumerate(dd_files_paths_in):
        if multimode:
            print(" ======== ", i + 1, "/", N, " ======== ")
            print(" === ", os.path.basename(dd_path))
        # names, dates etc ...
        Table_dd = pd.read_table(dd_path, sep=r"\s+", header=-1)

        dd_name = os.path.basename(dd_path)

        dd_start_jjul = np.min(Table_dd[12])
        dd_start_sec = np.min(Table_dd[13])
        dd_end_jjul = np.max(Table_dd[12])
        dd_end_sec = np.max(Table_dd[13])

        dd_start_epoch = conv.jjul_cnes2dt(dd_start_jjul) + dt.timedelta(
            seconds=dd_start_sec
        )
        dd_end_epoch = conv.jjul_cnes2dt(dd_end_jjul) + dt.timedelta(seconds=dd_end_sec)

        freq_dd_true = np.min(np.abs(np.diff(Table_dd[13])))

        # stat_A_lower = np.unique(Table_dd[10])[0]
        # stat_A = stat_A_lower.upper()
        # stat_B_lower = np.unique(Table_dd[11])[0]
        # stat_B = stat_B_lower.upper()

        stats_lower_list = list(set(list(Table_dd[10]) + list(Table_dd[11])))
        stats_list = [st.upper() for st in stats_lower_list]

        gins_path = get_gins_path()

        # be sure there is a TEMP DATA folder
        if temp_data_folder == "":
            temp_data_folder = os.path.join(gins_path, "gin", "TEMP_DATA")
        if not os.path.exists(temp_data_folder):
            os.makedirs(temp_data_folder)

        # be sure the RINEX is in gins folder ...
        bool_rinex_in_gin = check_if_file_in_gin_folder(dd_path)
        # ... and copy it otherwise
        if not bool_rinex_in_gin:
            print("INFO : will be copied in ", temp_data_folder)
            dd_path = copy_file_in_gin_folder(dd_path, temp_data_folder)

        if out_director_folder == "":
            out_director_folder = os.path.join(
                get_gins_path(), "gin", "data", "directeur"
            )

        if perso_orbclk:
            calccntr_suffix = "_" + ac + str(repro)
        else:
            calccntr_suffix = ""

        # coord type suffix
        if out_coords.upper() in ("FLH", "XYZ"):
            coord_prefix = "_" + out_coords.lower()
        else:
            coord_prefix = ""

        stats_str = "_".join(stats_list)
        director_output_name = (
            director_name_prefix
            + "_"
            + stats_str
            + "_"
            + str(dd_start_jjul)
            + "_"
            + str(dd_start_epoch.year)
            + "_"
            + conv.dt2doy(dd_start_epoch)
            + "_"
            + str(freq_dd_true)
            + coord_prefix
            + calccntr_suffix
            + ".yml"
        )

        director_output_path = os.path.join(out_director_folder, director_output_name)
        director_generik_file = open(director_generik_path)
        director_dic = yaml.load(director_generik_file)
        director_dic_proto = copy.deepcopy(director_dic)

        # ADDING RINEX NAME & DATES INTO THE DIR
        # name
        dd_path_dir_compatible = make_path_ginsstyle(dd_path)
        director_dic["observation"]["interobject_data"][1][
            "file"
        ] = dd_path_dir_compatible

        # date
        strt_day, strt_sec = conv.dt2jjul_cnes(dd_start_epoch, False)
        end_day, end_sec = conv.dt2jjul_cnes(dd_end_epoch, False)

        director_dic["date"]["arc_start"][0] = strt_day
        director_dic["date"]["arc_stop"][0] = end_day
        if "initial_state_vector_date" in list(director_dic["date"].keys()):
            director_dic["date"]["initial_state_vector_date"][0] = strt_day

        director_dic["date"]["arc_start"][1] = float(strt_sec)
        director_dic["date"]["arc_stop"][1] = float(end_sec)
        if "initial_state_vector_date" in list(director_dic["date"].keys()):
            director_dic["date"]["initial_state_vector_date"][1] = float(strt_sec)

        # interval
        if auto_interval:  # dir is modified only if freq_rnx >= 0
            director_dic["observation"]["removal"]["simulation_stepsize"] = int(
                freq_dd_true
            )
            # check if its not a static dir
            if (
                director_dic["parameter"]["adjustment_parameters"]["stations"][
                    "adjustment_frequency"
                ][-1]
                != 0
            ):
                director_dic["parameter"]["adjustment_parameters"]["stations"][
                    "adjustment_frequency"
                ] = [0, 0, 0, freq_dd_true]
        print(
            "INFO : auto_interval",
            auto_interval,
            ", simulation_stepsize : ",
            director_dic["observation"]["removal"]["simulation_stepsize"],
        )
        # output coords
        if out_coords.upper() in ("FLH", "PLH"):
            director_dic["parameter"]["adjustment_parameters"]["stations"][
                "coordinates"
            ] = "latitude_longitude_height"
        elif out_coords.upper() == "XYZ":
            director_dic["parameter"]["adjustment_parameters"]["stations"][
                "coordinates"
            ] = "cartesian_xyz"

        # =========   ORBITS   =============
        if not perso_orbclk:
            # GRG or GR2 for clock and orbits
            if dt.datetime(1998, 1, 4) <= dd_start_epoch <= dt.datetime(2013, 12, 29):
                gr_gr2 = "GR2"
            #    elif dt_rinex < dt.datetime(1998,1,4):
            #        gr_gr2 = 'GR1' # obsolete parce que GR1 intégralement inclu dans GR2
            else:
                gr_gr2 = "GRG"
            horpath = os.path.join("horloges", gr_gr2, "defaut")
            orbpath = os.path.join("orbites", gr_gr2, "defaut")
        else:
            orbpath, horpath = download_convert_2_gins_orb_clk(
                dd_start_epoch, temp_data_folder, ac=ac, repro=repro
            )

            # Exception case where no sp3/clk was found
            if orbpath == None or horpath == None:
                print(
                    "WARN : something went wrong with orbits/clocks download/conversion "  )
                print("       skiping the day ...")
                failed_dd_date_lis.append(dd_start_epoch)
                if multimode:
                    continue
                else:
                    return ""

            orbpath = make_path_ginsstyle(orbpath)
            horpath = make_path_ginsstyle(horpath)

        director_dic["model"]["environment"]["gnss_clock"] = horpath
        director_dic["observation"]["interobject_data"][0]["file"] = orbpath

        # correcting a pyyaml bug : version is interpreted as a int instead of a string
        verslis = list(str(director_dic["version"]))

        verstr = verslis[0] + verslis[1] + "_" + "_".join(verslis[2:])
        director_dic["version"] = verstr

        # Security CHECKs
        statfile_path_ginsstyle = director_dic["object"]["station"][
            "station_coordinates"
        ]
        statfile_path_full = os.path.join(gins_path, statfile_path_ginsstyle[6:])

        ocloadfile_path_ginsstyle = director_dic["object"]["station"][
            "ocean_tide_loading"
        ]
        ocloadfile_path_full = os.path.join(gins_path, ocloadfile_path_ginsstyle[6:])

        # check_stat_in_statfile(stat,statfile_path_full)

        # domes = find_domes_in_statfile(stat,statfile_path_full)
        # print 'INFO : DOMES : ',domes
        # chek_domes_oclo(domes[0],ocloadfile_path_full)

        # Case of kinematic process : need to change keys in user extension
        stat_B = None
        if "user_extension" in director_dic:
            print("INFO : kinematic process with 'user_extension' fields")
            if "userext_gps__haute_freq" in director_dic["user_extension"]:
                director_dic["user_extension"]["userext_gps__haute_freq"] = stat_B
            if "userext_gps__haute_freq".upper() in director_dic["user_extension"]:
                director_dic["user_extension"][
                    "userext_gps__haute_freq".upper()
                ] = stat_B

            if "userext_addition" in director_dic["user_extension"]:
                if (
                    "gps__haute_freq"
                    in director_dic["user_extension"]["userext_addition"]
                ):
                    director_dic["user_extension"]["userext_addition"][
                        "gps__haute_freq"
                    ] = stat_B
                if (
                    "gps__haute_freq".upper()
                    in director_dic["user_extension"]["userext_addition"]
                ):
                    director_dic["user_extension"]["userext_addition"][
                        "gps__haute_freq".upper()
                    ] = stat_B

        if "userext_addition" in director_dic["user_extension"]:
            if np.any(
                [
                    "gps__haute_freq".upper() in e
                    for e in director_dic["user_extension"]["userext_addition"]
                ]
            ):
                print("INFO : kinematic process with 'userext_addition' fields")
                index_gps_hf = [
                    "gps__haute_freq".upper() in e
                    for e in director_dic["user_extension"]["userext_addition"]
                ].index(True)
                director_dic["user_extension"]["userext_addition"][index_gps_hf] = (
                    "gps__haute_freq ".upper() + stat_B
                )

        # return director_dic_proto , director_dic

        # WRITING THE NEW DIRECTOR
        print("INFO : writing : ", director_output_path)
        with open(director_output_path, "w+") as outfile:
            outfile.write(yaml.dump(director_dic, default_flow_style=False))

        # correcting a bug : GINS YAML interpreter can't manage false or true ...
        utils.replace(director_output_path, "false", "no")
        utils.replace(director_output_path, "true", "yes")

        director_output_path_lis.append(director_output_path)

    if multimode:
        if len(failed_dd_date_lis) > 0:
            print("INFO : ", len(failed_dd_date_lis), "/", N, " FILEs failed")
            print("       ", [str(d) for d in failed_dd_date_lis])
        return director_output_path_lis
    else:
        return director_output_path


def double90_runner(inp_prarie_file, temp_data_folder=""):

    # be sure there is a TEMP DATA folder
    if temp_data_folder == "":
        temp_data_folder = os.path.join(get_gins_path(1), "TEMP_DATA")
    if not os.path.exists(temp_data_folder):
        os.makedirs(temp_data_folder)

    curdir = os.getcwd()
    os.chdir(temp_data_folder)

    command = "double90 " + inp_prarie_file
    print("INFO : double90 command lanched :")
    print(command)
    print("")

    p = subprocess.Popen(
        [command], shell=True
    )  # , stdout=subprocess.PIPE , stderr=subprocess.PIPE)
    p.wait()
    outname = os.path.basename(inp_prarie_file).split(".")[0] + ".dd"
    os.rename("./fort.20", outname)
    os.chdir(curdir)

    outpath = os.path.join(temp_data_folder, outname)

    return outpath


def double_diff_binom(
    rinex_path_A, rinex_path_B, temp_data_folder="", final_data_folder=""
):
    """
    DISCONTINUED
    """

    # be sure there is a TEMP DATA folder
    if temp_data_folder == "":
        temp_data_folder = os.path.join(get_gins_path(1), "TEMP_DATA")
    if not os.path.exists(temp_data_folder):
        os.makedirs(temp_data_folder)

    statA = os.path.basename(rinex_path_A)[0:4]
    statB = os.path.basename(rinex_path_B)[0:4]

    strt_epochA, end_epochA, freq_rnx_trueA = operational.rinex_start_end(
        rinex_path_A, True
    )
    strt_epochB, end_epochB, freq_rnx_trueB = operational.rinex_start_end(
        rinex_path_B, True
    )

    year, doy = conv.dt2doy_year(strt_epochA)

    argdict = dict()
    # argdict['-ignore']  = 'EG'

    praA_path = prairie_manual(
        rinex_path_A, temp_data_folder=temp_data_folder, argsdict=argdict
    )
    praB_path = prairie_manual(
        rinex_path_B, temp_data_folder=temp_data_folder, argsdict=argdict
    )

    praABcat_name = utils.join_improved("_", statA, statB, year, doy) + ".pra.cat"
    praABcat_path = os.path.join(temp_data_folder, praABcat_name)

    utils.cat(praABcat_path, praA_path, praB_path)

    dd_path = double90_runner(praABcat_path, temp_data_folder=temp_data_folder)

    if final_data_folder != "":
        shutil.move(dd_path, final_data_folder)
        dd_path = os.path.join(final_data_folder, os.path.basename(dd_path))

    return dd_path


def double_diff_multi(
    rinex_path_list,
    temp_data_folder="",
    final_data_folder="",
    remove=True,
    ignore_glo_gal=False,
):
    """
    rinex_path_list :
        List of RINEXs paths

    temp_data_folder & final_data_folder :
        if empty string (''), files are saved in pygins_runner "TEMP_DATA" folder
        and in the specified folder else

    """

    # be sure there is a TEMP DATA folder
    if temp_data_folder == "":
        temp_data_folder = os.path.join(get_gins_path(1), "TEMP_DATA")
    if not os.path.exists(temp_data_folder):
        os.makedirs(temp_data_folder)

    prarie_path_list = []
    stats_list = []

    for rinex_path in rinex_path_list:

        print("****** conversion to GINS (PRARIE) format of :")
        print(rinex_path)

        bool_comp_rnx = operational.check_if_compressed_rinex(rinex_path)
        # ... if not crz2rnx !
        if bool_comp_rnx:
            crinex_path = rinex_path
            rinex_path = operational.crz2rnx(crinex_path, temp_data_folder)
            if not os.path.isfile(rinex_path):
                print("ERR : something went wrong during CRZ2RNX, skiping the file")
                print(crinex_path)
                continue

        statA = os.path.basename(rinex_path)[0:4]

        strt_epochA, end_epochA, freq_rnx_trueA = operational.rinex_start_end(
            rinex_path, True
        )
        doy, year = conv.dt2doy_year(strt_epochA)

        argdict = dict()

        if ignore_glo_gal:
            argdict["-ignore"] = "EG"

        praA_path = prairie_manual(
            rinex_path, temp_data_folder=temp_data_folder, argsdict=argdict
        )

        prarie_path_list.append(praA_path)
        stats_list.append(statA)

        if bool_comp_rnx:
            os.remove(rinex_path)

    pracat_name = (
        utils.join_improved("_", year, doy, *list(set(stats_list))) + ".pra.cat"
    )
    pracat_path = os.path.join(temp_data_folder, pracat_name)

    print("****** Concatenation of the PRAIRIE files in :")
    print(pracat_path)

    utils.cat(pracat_path, *prarie_path_list)

    print("****** generation of the Double Diff. file :")

    dd_path = double90_runner(pracat_path, temp_data_folder=temp_data_folder)

    print("double diff. file generated in :")
    print(dd_path)

    if final_data_folder != "":
        shutil.move(dd_path, final_data_folder)
        dd_path = os.path.join(final_data_folder, os.path.basename(dd_path))

    if remove:
        os.remove(pracat_path)
        [os.remove(e) for e in prarie_path_list]

    return dd_path


# ==========================================================================
# ==========================================================================


# def check_if_director_path_is_ok(director_path,gins_path=get_gin_path()):
# USELESS FOR THE MOMENT
#    # 2 choices : the director_path is absolute or relative
#    director_name = os.path.basename(director_path)
#    nominal_director_path =  os.path.join(gins_path,'gin/data/directeur',director_name)
#
#    if os.path.isabs(adir):
#        out = True
#
#        if not filecmp.cmp(nominal_director_path , director_path):
#            print 'WARN : nominal_director_path != director_path !!!'
#            print director_path
#            print nominal_director_path
#            out = False
#    else:
#        if os.path.isfile(nominal_director_path):
#            out = True
#        else:
#            print 'WARN : director_path dont exists !!!'
#            out = False
#    return out



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 13 15:11:12 2020

@author: psakicki

This sub-module of geodezyx.conv deals with rotation, implemented with matrices.

it can be imported directly with:
from geodezyx import conv

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

#### Import the logger
import logging

########## BEGIN IMPORT ##########
#### External modules
import numpy as np

#### geodeZYX modules
from geodezyx import utils

log = logging.getLogger('geodezyx')


##########  END IMPORT  ##########

#  _____       _        _   _               __  __       _        _               
# |  __ \     | |      | | (_)             |  \/  |     | |      (_)              
# | |__) |___ | |_ __ _| |_ _  ___  _ __   | \  / | __ _| |_ _ __ _  ___ ___  ___ 
# |  _  // _ \| __/ _` | __| |/ _ \| '_ \  | |\/| |/ _` | __| '__| |/ __/ _ \/ __|
# | | \ \ (_) | || (_| | |_| | (_) | | | | | |  | | (_| | |_| |  | | (_|  __/\__ \
# |_|  \_\___/ \__\__,_|\__|_|\___/|_| |_| |_|  |_|\__,_|\__|_|  |_|\___\___||___/
#                                                                                 

### Rotation matrix

### The rotation matrices conventions are described in details in 
### Sakic (2016, PhD manuscript) p126

### It is based on the book of Grewal, M. S., Weill, L. R., & Andrews, A. p. (2007).
### Global Positioning Systems, Inertial Navigation, and Integration. 
### Hoboken, NJ, USA : John Wiley and Sons, Inc.

def c_2d(theta, angtype='deg'):
    if angtype == 'deg':
        theta = np.deg2rad(theta)

    return np.array([[np.cos(theta), -np.sin(theta)],
                     [np.sin(theta), np.cos(theta)]])


def c_ned2enu():
    return np.array([[0, 1, 0], [1, 0, 0], [0, 0, -1]])


def c_enu2ned():
    return np.array([[0, 1, 0], [1, 0, 0], [0, 0, -1]])


def c_ned2ecef(phi, llambda, angtype='deg'):
    if angtype == 'deg':
        phi = np.deg2rad(phi)
        llambda = np.deg2rad(llambda)

    in_ned = np.array([1, 0, 0])
    ie_ned = np.array([0, 1, 0])
    id_ned = np.array([0, 0, 1])

    ix_ned = np.array([-np.cos(llambda) * np.sin(phi), -np.sin(llambda), -np.cos(llambda) * np.cos(phi)])
    iy_ned = np.array([-np.sin(llambda) * np.sin(phi), np.cos(llambda), -np.sin(llambda) * np.cos(phi)])
    iz_ned = np.array([np.cos(phi), 0, -np.sin(phi)])

    c_ned2ecef_out = np.array([[np.dot(ix_ned, in_ned), np.dot(ix_ned, ie_ned), np.dot(ix_ned, id_ned)],
                               [np.dot(iy_ned, in_ned), np.dot(iy_ned, ie_ned), np.dot(iy_ned, id_ned)],
                               [np.dot(iz_ned, in_ned), np.dot(iz_ned, ie_ned), np.dot(iz_ned, id_ned)]])

    return c_ned2ecef_out


def c_ecef2ned(phi, llambda, angtype='deg'):
    return np.linalg.inv(c_ned2ecef(phi, llambda, angtype))


def c_enu2ecef(phi, llambda, angtype='deg'):
    if angtype == 'deg':
        phi = np.deg2rad(phi)
        llambda = np.deg2rad(llambda)

    ie_enu = np.array([1, 0, 0])
    in_enu = np.array([0, 1, 0])
    iu_enu = np.array([0, 0, 1])

    ix_enu = np.array([-np.sin(llambda), -np.cos(llambda) * np.sin(phi), np.cos(llambda) * np.cos(phi)])
    iy_enu = np.array([np.cos(llambda), -np.sin(llambda) * np.sin(phi), np.sin(llambda) * np.cos(phi)])
    iz_enu = np.array([0, np.cos(phi), np.sin(phi)])

    c_enu2ecef_out = np.array([[np.dot(ix_enu, ie_enu), np.dot(ix_enu, in_enu), np.dot(ix_enu, iu_enu)],
                               [np.dot(iy_enu, ie_enu), np.dot(iy_enu, in_enu), np.dot(iy_enu, iu_enu)],
                               [np.dot(iz_enu, ie_enu), np.dot(iz_enu, in_enu), np.dot(iz_enu, iu_enu)]])

    return c_enu2ecef_out


def c_ecef2enu(phi, llambda, angtype='deg'):
    return c_enu2ecef(phi, llambda, angtype).T


def c_ecef2enu_sigma(phi, llambda, angtype='deg'):
    """
    Gives the transformation matrix between ECEF and ENU (East North Up) 
    ref. frame for the sigmas (variance propagation)


    Source
    ------
    https://stackoverflow.com/questions/51162460/converting-ecef-xyz-covariance-matrix-to-enu-covariance-matrix
    """
    if angtype == 'deg':
        phi = np.deg2rad(phi)
        llambda = np.deg2rad(llambda)


    c_ecef2enu_sigma_out = np.array([[- np.sin(llambda), np.cos(llambda), 0],
                                     [-np.sin(phi) * np.cos(llambda), -np.sin(phi) * np.sin(llambda), np.cos(phi)],
                                     [np.cos(phi) * np.cos(llambda), np.cos(phi) * np.sin(llambda), np.sin(phi)]])

    return c_ecef2enu_sigma_out


def c_rpy2enu(roll, pitch, yaw, angtype='deg'):
    if angtype == 'deg':
        roll = np.deg2rad(roll)
        pitch = np.deg2rad(pitch)
        yaw = np.deg2rad(yaw)

    r = roll
    p = pitch
    y = yaw

    ir = np.array([np.sin(y) * np.cos(p), np.cos(y) * np.cos(p), np.sin(p)])
    ip = np.array([np.cos(r) * np.cos(y) + np.sin(r) * np.sin(y) * np.sin(p),
                   -np.cos(r) * np.sin(y) + np.sin(r) * np.cos(y) * np.sin(p),
                   -np.sin(r) * np.cos(p)])
    iy = np.array([-np.sin(r) * np.cos(y) + np.cos(r) * np.sin(y) * np.sin(p),
                   np.sin(r) * np.sin(y) + np.cos(r) * np.cos(y) * np.sin(p),
                   - np.cos(r) * np.cos(p)])

    #    Ie = np.array([ np.sin(Y) * np.cos(p) ,
    #    np.cos(R) * np.cos(Y) + np.sin(R) * np.sin(Y) * np.sin(p) ,
    #    -np.sin(R) * np.cos(Y) + np.cos(R) * np.sin(Y) * np.sin(p) ])
    #
    #    In = np.array([ np.cos(Y) * np.cos(p),
    #     -np.cos(R) * np.sin(Y) + np.sin(R) * np.cos(Y) * np.sin(p),
    #     np.sin(R) * np.sin(Y) + np.cos(R) * np.cos(Y) * np.sin(p) ])
    #
    #    Iu = np.array([ np.sin(p),
    #     -np.sin(R) * np.cos(p),
    #     -np.cos(R) * np.cos(p) ])

    c_rpy2enu_out = np.hstack((ir[np.newaxis].T, ip[np.newaxis].T, iy[np.newaxis].T))

    return c_rpy2enu_out


def c_rpy2enu2(roll, pitch, yaw, angtype='deg'):
    if angtype == 'deg':
        roll = np.deg2rad(roll)
        pitch = np.deg2rad(pitch)
        yaw = np.deg2rad(yaw)

    r = roll
    p = pitch
    y = yaw

    sr = np.sin(r)
    cr = np.cos(r)
    sp = np.sin(p)
    cp = np.cos(p)
    sy = np.sin(y)
    cy = np.cos(y)

    c_rpy2enu_out = np.array([[sy * cp, cr * cy + sr * sy * sp, - sr * cy + cr * sy * sp],
                              [cy * cp, -cr * sy + sr * cy * sp, sr * sy + cr * cy * sp],
                              [sp, -sr * cp, -cr * cp]])

    return c_rpy2enu_out


def c_eci2rtn(p, v):
    """
    gives the transformation matrix between ECI ref. frame and RTN (aka RIC or RSW)

    Parameters
    ----------
    p : numpy.array
        3D vector, position of the object in ECI frame

    v : numpy.array
        3D vector, velocity of the object in ECI frame
        
    Returns
    -------
    C_eci2rtn : numpy.array
        transformation matrix
            
    References
    ----------
    "Coordinate Systems", ASEN 3200 1/24/06 George H. Born
    https://www.colorado.edu/ASEN/asen3200/handouts/Coordinate%20System.pdf
    """

    h = np.cross(p, v)

    r = p / np.linalg.norm(p)
    c = h / np.linalg.norm(h)
    i = np.cross(c, r)

    c_eci2rtn_out = np.column_stack((r, i, c))

    return c_eci2rtn_out


def c_rtn2rpy():
    c_rtn2rpy_out = np.array([[0., 0, -1.],
                              [1., 0., 0.],
                              [0, -1., 0.]])
    return c_rtn2rpy_out


def c_cep2itrs(xpole, ypole):
    """
    xpole and ypole are given in mas, miliarcsecond 
    
    References
    ----------
        Hofmann-Wellenhof,et al. 2008 - GNSS - p21
        Xu - 2007 - GPS - p17
        Xu - Orbits - p15
    """

    xpole2 = np.deg2rad(xpole * (1. / 3600.) * (10 ** -3))
    ypole2 = np.deg2rad(ypole * (1. / 3600.) * (10 ** -3))

    c_cep2itrs_out = np.array([[1., 0., xpole2],
                               [0., 1., -ypole2],
                               [-xpole2, ypole2, 1.]])

    return c_cep2itrs_out


def c_euler(phi, theta, psi):
    """
    Gives the matrix of an Euler rotation
        
    References
    ----------
        https://fr.wikipedia.org/wiki/Angles_d%27Euler
    """
    cphi = np.cos(phi)
    ctheta = np.cos(theta)
    cpsi = np.cos(psi)
    sphi = np.sin(phi)
    stheta = np.sin(theta)
    spsi = np.sin(psi)

    c_euler_out = np.array([[cpsi * cphi - spsi * ctheta * sphi, -cpsi * sphi - spsi * ctheta * cphi, spsi * stheta],
                            [spsi * cphi + cpsi * ctheta * sphi, -spsi * sphi + cpsi * ctheta * cphi, -cpsi * stheta],
                            [stheta * sphi, stheta * cphi, ctheta]])

    return c_euler_out


def c_x(theta):
    """
    Gives the rotation matrix along the X-axis
    
    Manage iterable (list, array) as input
    
    [1,0, 0]
    [0,C,-S]
    [0,S, C]
    
    References
    ----------
        https://fr.wikipedia.org/wiki/Matrice_de_rotation#En_dimension_trois
        
    """

    if not utils.is_iterable(theta):
        theta = np.array([theta])

    c = np.cos(theta)
    s = np.sin(theta)
    z = np.zeros(len(theta))
    i = np.ones(len(theta))

    c_x_out = np.stack([[i, z, z],
                        [z, c, -s],
                        [z, s, c]])

    c_x_out = np.squeeze(c_x_out)

    return c_x_out


def c_y(theta):
    """
    Gives the rotation matrix around the Y-axis
    
    Manage iterable (list, array) as input
    
    [ C,0,S]
    [ 0,1,0]
    [-S,0,C]
    
    References
    ----------
        https://fr.wikipedia.org/wiki/Matrice_de_rotation#En_dimension_trois
    """

    if not utils.is_iterable(theta):
        theta = np.array([theta])

    c = np.cos(theta)
    s = np.sin(theta)
    z = np.zeros(len(theta))
    i = np.ones(len(theta))

    c_y_out = np.stack([[c, z, s],
                        [z, i, z],
                        [-s, z, c]])

    c_y_out = np.squeeze(c_y_out)

    return c_y_out


def c_z(theta):
    """
    Gives the rotation matrix around the Z-axis

    [C,-S,0]
    [S, C,0]
    [0, 0,1]
    
    References
    ----------
        https://fr.wikipedia.org/wiki/Matrice_de_rotation#En_dimension_trois
    """

    if not utils.is_iterable(theta):
        theta = np.array([theta])

    c = np.cos(theta)
    s = np.sin(theta)
    z = np.zeros(len(theta))
    i = np.ones(len(theta))

    c_z_out = np.stack([[c, -s, z],
                        [s, c, z],
                        [z, z, i]])

    c_z_out = np.squeeze(c_z_out)

    return c_z_out


def rot_quelconq(theta, vx, vy, vz, angtype='deg'):
    n = np.linalg.norm([vx, vy, vz])

    vx = vx / n
    vy = vy / n
    vz = vz / n

    if angtype == 'deg':
        theta = np.deg2rad(theta)

    c = np.cos(theta)
    s = np.sin(theta)

    r = np.array([[vx ** 2 + (1 - vx ** 2) * c, vx * vy * (1 - c) - vz * s, vx * vz * (1 - c) + vy * s],
                  [vx * vy * (1 - c) + vz * s, vy ** 2 + (1 - vy ** 2) * c, vy * vz * (1 - c) - vx * s],
                  [vx * vz * (1 - c) - vy * s, vy * vz * (1 - c) + vx * s, vz ** 2 + (1 - vz ** 2) * c]])

    return r


def vector_rpy(V, out_deg=True, ad_hoc_mode=False):
    """
    Donne le "Roll Pitch Yaw" pour un Point/Vecteur
    (les coordonnées spheriques plus exactement, le roll sera tjrs nul)

    le ad_hoc_mode est la pour corriger un bug que je n'explique pas
    pour l'utilisation de ads_offset
    (cette methode a été trouvé par les scripts ENURPY2 et ENURPY3)
    """

    x = V[0]
    y = V[1]
    z = V[2]

    n = np.linalg.norm(V)

    roll = 0
    pitch = np.arcsin(z / n)
    yaw = np.arctan2(y, x)

    if out_deg:
        roll = np.rad2deg(roll)
        pitch = np.rad2deg(pitch)
        yaw = np.rad2deg(yaw)

    if ad_hoc_mode:
        roll = -roll
        pitch = -pitch
        yaw = -yaw

        roll, pitch, yaw = pitch, roll, yaw

    return roll, pitch, yaw


def add_offset(direction, delta, point=None,
               out_delta_enu=False):
    """
    Un mode adhoc a du être implémenté dans la fonction élémentaire
    vector_RPY pour que ça puisse marcher correctement, reste à
    comprendre pourquoi ...

    out_delta_enu est prioritaire sur  les autres modes de sortie
    et renvoie le  delta

    par défaut on a le delta + le vecteur Direction
    et si Point est précisé : Point + Direction
    """

    rpy = vector_rpy(direction, ad_hoc_mode=True)

    crpy2enu = c_rpy2enu(*rpy)
    # Cenu2rpy = np.linalg.inv(C_rpy2enu(*rpy))

    delta_enu = crpy2enu.dot(c_enu2ned().dot(delta))

    if out_delta_enu:
        log.info(out_delta_enu)

    if not np.isclose(np.linalg.norm(delta), np.linalg.norm(delta_enu)):
        log.warning("np.linalg.norm(Delta) != np.linalg.norm(Delta_enu)")
        log.warning(np.linalg.norm(delta), np.linalg.norm(delta_enu), \
                    np.linalg.norm(delta) - np.linalg.norm(delta_enu))

    if not type(point) is None:
        outpoint = point + delta_enu
    else:
        outpoint = direction + delta_enu

    return outpoint



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# https://chatgpt.com/c/683e9336-2858-8012-b4a9-3c9308c0bb46

"""
Created on Tue Jun  3 08:23:57 2025

@author: psakicki

"""


from netCDF4 import Dataset, date2num
import numpy as np
import datetime as dt

# Create NetCDF file
ncfile = Dataset('example_oceansites.nc', mode='w', format='NETCDF4_CLASSIC')

# Global attributes (OceanSITES standard)
ncfile.setncatts({
    'title': 'Example OceanSITES Time Series',
    'institution': 'Your Organization',
    'source': 'Mooring',
    'history': 'Created ' + dt.datetime.now(dt.UTC).strftime('%Y-%m-%d %H:%M:%S UTC'),
    'references': 'http://www.oceansites.org',
    'Conventions': 'CF-1.6 OceanSITES-1.3',
    'netcdf_version': 'NETCDF4_CLASSIC',
    'deployment_code': 'example01',
    'id': 'EXAMPLE_001',
    'data_type': 'OceanSITES time-series data',
    'format_version': '1.3',
    'site_code': 'EXAMPLE',
    'platform_code': 'EX01',
    'instrument_number': '001',
    'data_mode': 'R'
})

# Define dimensions
ncfile.createDimension('TIME', None)   # unlimited
ncfile.createDimension('DEPTH', 1)
ncfile.createDimension('LATITUDE', 1)
ncfile.createDimension('LONGITUDE', 1)

# Create variables
times = ncfile.createVariable('TIME', 'f8', ('TIME',))
times.units = 'days since 1950-01-01 00:00:00 UTC'
times.long_name = 'Time'
times.standard_name = 'time'
times.axis = 'T'

depths = ncfile.createVariable('DEPTH', 'f4', ('DEPTH',))
depths.units = 'm'
depths.positive = 'down'
depths.axis = 'Z'
depths.long_name = 'Depth'
depths.standard_name = 'depth'

lat = ncfile.createVariable('LATITUDE', 'f4', ('LATITUDE',))
lat.units = 'degrees_north'
lat.long_name = 'Latitude'
lat.standard_name = 'latitude'
lat.axis = 'Y'

lon = ncfile.createVariable('LONGITUDE', 'f4', ('LONGITUDE',))
lon.units = 'degrees_east'
lon.long_name = 'Longitude'
lon.standard_name = 'longitude'
lon.axis = 'X'

# Example data variable (e.g., temperature)
temp = ncfile.createVariable('TEMP', 'f4', 
                             ('TIME', 'DEPTH', 'LATITUDE', 'LONGITUDE'),
                             fill_value=-9999.0)
temp.units = 'degree_Celsius'
temp.long_name = 'Sea Water Temperature'
temp.standard_name = 'sea_water_temperature'
temp.coordinates = 'TIME DEPTH LATITUDE LONGITUDE'

# Fill dimensions with example data
num_points = 10
start_time = dt.datetime(2023, 1, 1)
time_vals = [start_time + dt.timedelta(hours=6*i) for i in range(num_points)]
times[:] = date2num(time_vals, units=times.units)

depths[:] = [10.0]
lat[:] = [45.0]
lon[:] = [-30.0]

# Add mock temperature data
temp[:, 0, 0, 0] = np.random.uniform(10, 15, size=num_points)

# Close the NetCDF file
ncfile.close()
print("OceanSITES NetCDF file created successfully.")





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#### IMPORT EXTERNAL MODULES
import logging.config # MUST remain logging.config for < v3.10
# (import logging alone doesn't work for < 3.10)
import os
from os import path

__version__='5.0.0'  ## increase it with bump-my-version !!!

#### IMPORT CONFIG FOR LOGGER
log_file_path = os.path.join(path.dirname(path.abspath(__file__)),'logconfig','loggzyx.py')
from . import logconfig
if os.path.isfile(log_file_path):
    ##print("INFO:",log_file_path,"found")
    ###### old config file format
    ###logging.config.fileConfig(fname=log_file_path, disable_existing_loggers=False)
    ###### new dict
    logging.config.dictConfig(logconfig.loggzyx.log_config_dict)
else:
    print("ERR:logger config file",log_file_path,"is missing")


#### IMPORT GEODEZYX INTERNAL SUBMODULES
# from . import externlib # disable 202406xx
# from . import megalib # disable 202411xx


from . import conv
from . import utils

#from . import athmo
#from . import files_rw
#from . import geodyn
#from . import interp
#from . import marine
#from . import operational
#from . import reffram
#from . import stats
#from . import time_series
#from . import utils_xtra

__all__ = ['athmo',
           'conv',
           'externlib',
           'files_rw',
           'geodyn',
           'interp',
           'marine',
           'operational',
           'reffram',
           'stats',
           'time_series',
           'utils',
           'utils_xtra']

if __name__ == '__main__':
    print("geodezyx version",__version__)
    print("geodezyx.__all__",__all__)

#### Import extern libraires in the geodezyx namespace
#from geodezyx.extern import *

# NB : if you want extern in the main namespace 
#      don't call with 
#      import geodezyx
#      but with
#      from geodezyx import *




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic
This sub-module of geodezyx.operational contains functions to download
gnss data and products from distant IGS servers.
it can be imported directly with:
from geodezyx import operational
The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License
Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

import ftplib

#### Import the logger
import logging
import os
import pathlib
import shutil
import time
import urllib

########## BEGIN IMPORT ##########
#### External modules
from ftplib import FTP, FTP_TLS
from multiprocessing.dummy import Pool as ThreadPool

import numpy as np
import pandas as pd

#### geodeZYX modules
from geodezyx import conv
from geodezyx import utils
from geodezyx.stats import outlier_mad

log = logging.getLogger("geodezyx")


##########  END IMPORT  ##########


def start_end_date_easy(start_year, start_doy, end_year, end_doy):
    """
    generates start/end datetimes from a start/end year/day of year

    Parameters
    ----------
    start_year : int
        start year.
    start_doy : int
        start day of year.
    end_year : int
        end year.
    end_doy : int
        end day of year.

    Returns
    -------
    start : datetime
        converted start datetime.
    end : datetime
        converted end datetime.

    """
    start = conv.doy2dt(start_year, start_doy)
    end = conv.doy2dt(end_year, end_doy)
    return start, end


def effective_save_dir_orbit(
    parent_archive_dir, calc_center, date, archtype="year/doy/"
):
    """
    INTERNAL_FUNCTION

    archtype =
        stat
        stat/year
        stat/year/doy
        year/doy
        year/stat
        week/dow
        wkwwww : use a GFZ's CF-ORB wk<wwww> naming
        OR only '/' for a dirty saving in the parent folder
        ... etc ...
    """
    if archtype == "/":
        return parent_archive_dir

    out_save_dir = parent_archive_dir
    fff = archtype.split("/")
    year = str(date.year)
    doy = conv.dt2doy(date)
    week, dow = conv.dt2gpstime(date)

    for f in fff:
        if "wkwwww" in f:
            f_evaluated = "wk" + str(week).zfill(4)
        else:
            f_evaluated = eval(f)
        out_save_dir = os.path.join(out_save_dir, str(f_evaluated))
    return out_save_dir


#  _    _ _______ _______ _____    _____                      _                 _
# | |  | |__   __|__   __|  __ \  |  __ \                    | |               | |
# | |__| |  | |     | |  | |__) | | |  | | _____      ___ __ | | ___   __ _  __| |
# |  __  |  | |     | |  |  ___/  | |  | |/ _ \ \ /\ / / '_ \| |/ _ \ / _` |/ _` |
# | |  | |  | |     | |  | |      | |__| | (_) \ v  v /| | | | | (_) | (_| | (_| |
# |_|  |_|  |_|     |_|  |_|      |_____/ \___/ \_/\_/ |_| |_|_|\___/ \__,_|\__,_|


#### HTTP classic Download


def downloader(
    url, savedir, force=False, check_if_file_already_exists_uncompressed=True
):
    """
    general function to download a file

    can also handle non secure FTP
    """

    if type(url) is tuple:
        need_auth = True
        username = url[1]
        password = url[2]
        url = url[0]
    else:
        need_auth = False
        username = ""
        password = ""

    url_print = str(url)

    rnxname = os.path.basename(url)

    pot_compress_files_list = [os.path.join(savedir, rnxname)]

    if check_if_file_already_exists_uncompressed:
        pot_compress_files_list.append(
            os.path.join(savedir, rnxname.replace(".gz", ""))
        )
        pot_compress_files_list.append(os.path.join(savedir, rnxname.replace(".Z", "")))
        pot_compress_files_list = list(set(pot_compress_files_list))

    for f in pot_compress_files_list:
        if os.path.isfile(f) and (not force):
            log.info(os.path.basename(f) + " already exists locally ;)")
            return None

    ##### LOCAL FILE (particular case for GFZ)
    if os.path.isfile(url):
        log.info("INFO : downloader : the is a local file, a simple copy will be used")
        log.info("       URL : %s", url)
        shutil.copy(url, savedir)

    ##### REMOTE FILE (General case)
    elif ("http" in url) or (("ftp" in url) and not need_auth):
        # managing an authentification
        if need_auth:  # HTTP with Auth
            password_mgr = urllib.request.HTTPPasswordMgrWithDefaultRealm()
            top_level_url = url
            password_mgr.add_password(None, top_level_url, username, password)
            handler = urllib.request.HTTPBasicAuthHandler(password_mgr)
            # create "opener" (OpenerDirector instance)
            opener = urllib.request.build_opener(handler)
        else:  # FTP or HTTP without Auth
            opener = urllib.request.build_opener()

        # use the opener to fetch a URL
        try:
            f = opener.open(url)
        except (urllib.error.HTTPError, urllib.error.URLError) as exp:
            log.warning("%s not found :(", rnxname)
            log.warning(url_print)
            log.warning(exp)
            return ""

        log.info("%s found, downloading :)", rnxname)
        data = f.read()
        if not os.path.exists(savedir):
            os.makedirs(savedir)
        outpath = os.path.join(savedir, rnxname)
        with open(outpath, "wb") as code:
            code.write(data)
        return_str = outpath

    elif ("ftp" in url) and need_auth:
        log.critical("MUST BE IMPEMENTED")
        return_str = ""
    else:
        log.error("something goes wrong with the URL")
        log.error(url)
        return_str = ""

    return return_str


def downloader_wrap(intup):
    downloader(*intup)
    return None


#  ______ _______ _____    _____                      _                 _
# |  ____|__   __|  __ \  |  __ \                    | |               | |
# | |__     | |  | |__) | | |  | | _____      ___ __ | | ___   __ _  __| |
# |  __|    | |  |  ___/  | |  | |/ _ \ \ /\ / / '_ \| |/ _ \ / _` |/ _` |
# | |       | |  | |      | |__| | (_) \ v  v /| | | | | (_) | (_| | (_| |
# |_|       |_|  |_|      |_____/ \___/ \_/\_/ |_| |_|_|\___/ \__,_|\__,_|


#### FTP DOWNLOAD


class MyFTP_TLS(FTP_TLS):
    """
    This class is a subclass of FTP_TLS from the ftplib module. It is used to create an FTPS client that shares the TLS session.
    This is to avoid the error: ssl.SSLEOFError: EOF occurred in violation of protocol (_ssl.c:2396)
    Source: https://stackoverflow.com/questions/14659154/ftps-with-python-ftplib-session-reuse-required

    Methods
    -------
    ntransfercmd(cmd, rest=None)
        Initiate a data transfer over a new connection.
    """

    def ntransfercmd(self, cmd, rest=None):
        """
        Initiate a data transfer over a new connection.

        Parameters
        ----------
        cmd : str
            The command to send to the server.
        rest : str, optional
            A string that contains a marker representing where the server is to restart the operation's data transfer. Default is None.

        Returns
        -------
        tuple
            The connection and the expected size of the data.

        Notes
        -----
        If the protection level is set to private (i.e., _prot_p is True), the connection is wrapped in an SSL/TLS layer.
        """
        conn, size = FTP.ntransfercmd(self, cmd, rest)
        if self._prot_p:
            conn = self.context.wrap_socket(
                conn, server_hostname=self.host, session=self.sock.session
            )  # this is the fix
        return conn, size


def ftp_dir_list_files(ftp_obj_in):
    """
    Lists the files in the current directory of the FTP object.

    Parameters
    ----------
    ftp_obj_in : FTP object
        The FTP object used to connect to the FTP server.

    Returns
    -------
    list
        A list of filenames in the current directory of the FTP object.

    Notes
    -----
    This function tries to get the list of filenames in the current directory of the FTP object.
    If it encounters a permission error, it checks if the error message is "550 No files found".
    If it is, it logs a warning message and returns an empty list.
    If the error message is different, it raises the exception.
    """
    files = []
    try:
        files = ftp_obj_in.nlst()
    except ftplib.error_perm as resp:
        if str(resp) == "550 No files found":
            log.warning("No files in this directory" + ftp_obj_in.pwd())
        else:
            raise
    return files


def ftp_objt_create(
    secure_ftp_inp=False,
    host="",
    chdir="",
    parallel_download=1,
    user="anonymous",
    passwd="",
    retry_count=3,
):
    """
    This function creates and returns an FTP object and a list of FTP objects for multiple downloads.

    Parameters
    ----------
    secure_ftp_inp : bool
        If True, uses FTPS for secure file transfer. Default is False.
    host : str, optional
        The hostname of the FTP server. Default is an empty string.
    chdir : str, optional
        The directory to change to after connecting to the FTP server. Default is an empty string.
    parallel_download : int, optional
        The number of parallel downloads to be performed. Default is 1.
    user : str, optional
        The username for the FTP server. Default is "anonymous".
    passwd : str, optional
        The password for the FTP server. Default is an empty string.
    retry_count : int, optional
        The number of times to retry creating the FTP object. Default is 3.

    Returns
    -------
    tuple
        The main FTP object for crawling and a list of FTP objects for parallel downloads.

    Notes
    -----
    This function creates an FTP object using the appropriate constructor based on the secure_ftp_inp parameter.
    It then creates a list of FTP objects for multiple downloads.
    If a directory is specified, it changes the current working directory of the main FTP object to that directory.
    """

    # define the right constructor
    if secure_ftp_inp:
        ftp_constuctor = MyFTP_TLS
    else:
        ftp_constuctor = FTP

    # create a list of FTP object for multiple downloads
    ftp_obj_list_out = []
    for i in range(parallel_download):
        for attempt in range(retry_count):
            try:
                current_ftp_obj = ftp_constuctor(host)
                ftp_obj_list_out.append(current_ftp_obj)
                break  # Exit the retry loop if successful
            except Exception as e:
                log.warning("FTP object creation failed on attempt %d", attempt + 1)
                log.warning(e)
                time.sleep(5)
                if attempt == retry_count - 1:
                    log.error("Max retries reached. Could not create FTP object.")

    if secure_ftp_inp:
        [f.login(user, passwd) for f in ftp_obj_list_out]
        [f.prot_p() for f in ftp_obj_list_out]
    else:
        [f.login(user=user, passwd=passwd) for f in ftp_obj_list_out]

    # define the main obj for crawling
    ftp_main = ftp_obj_list_out[0]

    # change the directory of the main ftp obj if we ask for it
    if chdir:
        log.info("Move to: %s", chdir)
        ftp_main.cwd(chdir)

    return ftp_main, ftp_obj_list_out


def ftp_downld_core(ftp_obj, filename, localdir, force=False):
    """
    Performs the FTP download if we are already in the correct FTP folder.
    This is an internal function of ftp_downld.

    Parameters
    ----------
    ftp_obj : FTP object
        The FTP object used to connect to the FTP server.
    filename : str
        The name of the file to be downloaded.
    localdir : str
        The local directory where the downloaded file should be saved.
    force : bool, optional
        If True, forces the download even if the file already exists locally.
        Default is False.

    Returns
    -------
    tuple
        The local path of the downloaded file and a boolean indicating whether the download was successful.

    Notes
    -----
    This function first checks if the local directory exists, if not it creates it.
    Then it checks if the file already exists locally, if it does, it logs a message and returns.
    If the file does not exist, it tries to download the file from the FTP server.
    If the download is successful, it logs a success message and returns the local path and True.
    If the download fails, it logs a failure message and returns the local path and False.
    """

    localpath = os.path.join(localdir, filename)

    if not os.path.isdir(localdir):
        utils.create_dir(localdir)

    dl_go = True
    # Check if the file already exists locally
    bool_dl = False
    if not utils.empty_file_check(localpath):
        if not force:
            log.info(filename + " already exists ;)")
            bool_dl = True
            dl_go = False
        else:
            log.info(filename + " already exists, but re-download forced")
            bool_dl = False
            dl_go = True

    if dl_go:
        try:
            localfile = open(localpath, "wb")
            ftp_obj.retrbinary("RETR " + filename, localfile.write, 1024)
            localfile.close()
            bool_dl = True
            log.info(filename + " downloaded :)")

        except Exception as e:
            log.warning(localpath + " download failed :(")
            log.warning(e)
            bool_dl = False

    return localpath, bool_dl


def ftp_downld_mono(ftp_obj, full_remote_path, localdir, force=False):
    """
    Downloads a file through FTP protocol.

    Parameters
    ----------
    ftp_obj : FTP object
        The FTP object used to connect to the FTP server.
    full_remote_path : str
        The full path of the file on the FTP server.
    localdir : str
        The local directory where the downloaded file should be saved.
    force : bool, optional
        If True, forces the download even if the file already exists locally.
        Default is False.

    Returns
    -------
    tuple
        The output of the ftp_downloader_core function.

    Notes
    -----
    This function changes the current working directory of the FTP object to the directory of the file to be downloaded,
    and then calls the ftp_downloader_core function to download the file.
    """

    filename = os.path.basename(full_remote_path)
    intermed_path = full_remote_path.split("/")[3:]
    intermed_path.remove(filename)
    intermed_path = "/" + "/".join(intermed_path)

    ftp_obj.cwd(intermed_path)

    return ftp_downld_core(ftp_obj, filename, localdir, force=force)


def ftp_downld_wrap(intup):
    """
    This function is a wrapper for the ftp_downld function. It unpacks the input tuple and passes it to the ftp_downld function.

    Parameters
    ----------
    intup : tuple
        A tuple containing the parameters to be passed to the ftp_downld function.

    Returns
    -------
    tuple
        The output of the ftp_downld function.
    """
    outtup = ftp_downld_mono(*intup)
    return outtup


def ftp_downld_front(
    urls,
    savedirs,
    parallel_download=1,
    secure_ftp=False,
    user="anonymous",
    passwd="anonymous@isp.com",
    force=True,
):
    """
    This function is used to download files from FTP servers in parallel.

    Parameters
    ----------
    urls : str or list
        The URL or list of URLs of the files to be downloaded.
    savedirs : str or list
        The directory or list of directories where the downloaded files should be saved.
    parallel_download : int, optional
        The number of parallel downloads to be performed. Default is 1.
    secure_ftp : bool, optional
        If True, uses FTPS for secure file transfer. Default is False.
    user : str, optional
        The username for the FTP server. Default is "anonymous".
    passwd : str, optional
        The password for the FTP server. Default is 'anonymous@isp.com'.
    force : bool, optional
        If True, forces the download even if the file already exists. Default is True.

    Returns
    -------
    out_tup_lis : List of tuples
        Returns a list of tuples containing
        the local path of the downloaded file and
        a boolean indicating whether the download was successful.
        e.g. [(local_path1, True), (local_path2, False), ...]

    Notes
    -----
    This function uses the ThreadPool for parallel downloads.
    """

    # Check if urls and savedirs are iterable, if not convert them to list
    urllist = urls if utils.is_iterable(urls) else [urls]
    savedirlist = savedirs if utils.is_iterable(savedirs) else [savedirs] * len(urllist)
    secure_ftp_use = secure_ftp[0] if utils.is_iterable(secure_ftp) else secure_ftp
    ##### dirty to select secure_ftp 1st elt only....

    # Check if the length of urllist and savedirlist are the same
    if len(urllist) != len(savedirlist):
        log.error(
            "URL & out dirs lists do not have the same length: %s %s",
            len(urllist),
            len(savedirlist),
        )

    # Extract the host from the urls
    urlpathobj = pd.Series(urllist).apply(pathlib.Path)
    host_use = urlpathobj.apply(lambda p: p.parts[1]).unique()[0]

    # Create the FTP object
    ftpobj_main, ftpobj_lis = ftp_objt_create(
        secure_ftp_inp=secure_ftp_use,
        host=host_use,
        parallel_download=parallel_download,
        user=user,
        passwd=passwd,
    )

    # Create a list of FTP objects for parallel downloads
    ftpobj_mp_lis = ftpobj_lis * int(np.ceil(len(urllist) / parallel_download))
    force_lis = [force] * len(urllist)

    # Check if there are less FTP objects than URLs for parallel download
    if len(ftpobj_mp_lis) < len(urllist):
        log.warning(
            "less FTP objects than URL for parallel download, contact the main developper"
        )

    # Create a ThreadPool for parallel downloads
    pool = ThreadPool(parallel_download)

    # Start the parallel downloads
    out_tup_lis = pool.map(
        ftp_downld_wrap, list(zip(ftpobj_mp_lis, urllist, savedirlist, force_lis))
    )

    return out_tup_lis


def ftp_downloader_wo_objects(tupin):
    """
    create the necessary FTP object

    should not be used anymore
    """
    arch_center_main, wwww_dir, filename, localdir = tupin
    ftp_obj_wk = FTP(arch_center_main)
    ftp_obj_wk.login()
    ftp_obj_wk.cwd(wwww_dir)
    localpath, bool_dl = ftp_downld_core(ftp_obj_wk, filename, localdir)
    ftp_obj_wk.close()
    return localpath, bool_dl


def ftp_files_crawler_legacy(urllist, savedirlist, secure_ftp):
    """
    filter urllist,savedirlist generated with download_gnss_rinex with an
    optimized FTP crawl

    """
    ### create a DataFrame based on the urllist and savedirlist lists
    df = pd.concat((pd.DataFrame(urllist), pd.DataFrame(savedirlist)), axis=1)
    df_orig = df.copy()

    ### rename the columns
    if df.shape[1] == 4:
        loginftp = True
        df.columns = ("url", "user", "pass", "savedir")
    else:
        loginftp = False
        df.columns = ("url", "savedir")
        df["user"] = "anonymous"
        df["pass"] = ""

    ### Do the correct split for the URLs
    df = df.sort_values("url")
    df["url"] = df["url"].str.replace("ftp://", "")
    df["dirname"] = df["url"].apply(os.path.dirname)
    df["basename"] = df["url"].apply(os.path.basename)
    df["root"] = [e.split("/")[0] for e in df["dirname"].values]
    df["dir"] = [e1.replace(e2, "")[1:] for (e1, e2) in zip(df["dirname"], df["root"])]
    df["bool"] = False

    #### Initialisation of the 1st variables for the loop
    prev_row_ftpobj = df.iloc[0]
    prev_row_cwd = df.iloc[0]
    ftp_files_list = []
    count_loop = 0  # restablish the connexion after 50 loops (avoid freezing)
    #### Initialisation of the FTP object

    ftpobj, _ = ftp_objt_create(
        secure_ftp_inp=secure_ftp,
        host=prev_row_ftpobj.root,
        user=prev_row_ftpobj.user,
        passwd=prev_row_ftpobj["pass"],
    )

    for irow, row in df.iterrows():
        count_loop += 1

        ####### we recreate a new FTP object if the root URL is not the same
        if row.root != prev_row_ftpobj.root or count_loop > 20:
            ftpobj, _ = ftp_objt_create(
                secure_ftp_inp=secure_ftp,
                host=prev_row_ftpobj.root,
                user=prev_row_ftpobj.user,
                passwd=prev_row_ftpobj["pass"],
            )

            prev_row_ftpobj = row
            count_loop = 0

        ####### we recreate a new file list if the date path is not the same
        if (prev_row_cwd.dir != row.dir) or irow == 0:
            log.info("chdir " + row.dirname)
            ftpobj.cwd("/")

            try:  #### we try to change for the right folder
                ftpobj.cwd(row.dir)
            except:  #### If not possible, then no file in the list
                ftp_files_list = []

            ftp_files_list = ftp_dir_list_files(ftpobj)
            prev_row_cwd = row

            ####### we check if the files is avaiable
        if row.basename in ftp_files_list:
            df.loc[irow, "bool"] = True
            log.info(row.basename + " found on server :)")
        else:
            df.loc[irow, "bool"] = False
            log.warning(row.basename + " not found on server :(")

    df_good = df[df["bool"]].copy()

    df_good["url"] = "ftp://" + df_good["url"]

    ### generate the outputs
    if loginftp:
        urllist_out = list(zip(df_good.url, df_good.user, df_good["pass"]))
    else:
        urllist_out = list(df_good.url)

    savedirlist_out = list(df_good.savedir)

    return urllist_out, savedirlist_out



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.operational contains functions to run the 
time series velocities estimation software HECTOR. 

it can be imported directly with:
from geodezyx import operational

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

########## BEGIN IMPORT ##########
#### External modules
import datetime as dt
import glob
#### Import the logger
import logging
import os
import shutil
import subprocess
import time

import matplotlib.pyplot as plt
import numpy as np

from geodezyx import reffram
#### geodeZYX modules
from geodezyx import utils

log = logging.getLogger('geodezyx')

##########  END IMPORT  ##########
#  _    _ ______ _____ _______ ____  _____
# | |  | |  ____/ ____|__   __/ __ \|  __ \
# | |__| | |__ | |       | | | |  | | |__) |
# |  __  |  __|| |       | | | |  | |  _  /
# | |  | | |___| |____   | | | |__| | | \ \
# |_|  |_|______\_____|  |_|  \____/|_|  \_\
#


def keeping_specific_stats(listoffiles, specific_stats, invert=False):
    """if invert = True : NOT keeping BUT removing specific stats"""
    if not (type(specific_stats) is tuple):
        raise Exception("keeping_specific_stats : specific_stats must be a tuple")
    if not invert:
        outlistoffiles = []
    else:
        outlistoffiles = list(listoffiles)

    if specific_stats != ():
        for fil in listoffiles:
            for stat in specific_stats:
                if stat in fil:
                    if not invert:
                        outlistoffiles.append(fil)
                    else:
                        outlistoffiles.remove(fil)
        return outlistoffiles
    else:
        return listoffiles


def neufile_outlier_removing(
    inp_neufile, generik_conf_file, outdir="", remove_ctl_file=True
):
    """from a NEU file
    => removeoutlier preprocessing
    => 3 MOM files (one per component)"""
    neufile_name = os.path.basename(inp_neufile)
    prefix_inp = neufile_name.replace(".neu", "_")
    inpdir = os.path.dirname(inp_neufile)
    if outdir == "":
        outdir = inpdir
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    comp_list = ["North", "East", "Up"]

    if utils.grep_boolean(inp_neufile, "Components : YXZ"):
        comp_list_filename = ["Y", "X", "Z"]
    else:
        comp_list_filename = ["N", "E", "U"]

    for comp, comp_fn in zip(comp_list, comp_list_filename):
        work_conf_file = inpdir + "/" + prefix_inp + comp_fn + "_rmoutlier.ctl"
        prefix_out = prefix_inp + comp_fn + "_pre"
        momfile_name = prefix_out + ".mom"
        momfile_path = outdir + "/" + momfile_name
        shutil.copyfile(generik_conf_file, work_conf_file)
        #        work_conf_file_obj = open(work_conf_file,'w+')
        utils.replace(work_conf_file, "DataFile", "DataFile            " + neufile_name)
        utils.replace(work_conf_file, "DataDirectory", "DataDirectory       " + inpdir)
        utils.replace(
            work_conf_file, "OutputFile", "OutputFile          " + momfile_path
        )
        utils.replace(work_conf_file, "component", "component           " + comp)

        p = subprocess.Popen(
            "",
            executable="/bin/bash",
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        command = "removeoutliers " + work_conf_file
        log.info("LAUNCHING : %s", command)
        stdout, stderr = p.communicate(command.encode())
        # logs files
        std_file = open(outdir + "/" + prefix_out + ".std.log", "w")
        std_file.write(stdout.decode("utf-8"))
        std_file.close()
        if stderr:
            log.warning("err.log is not empty, must be checked !")
            log.warning(stderr.decode("utf-8"))
            log.warning("")
            err_file = open(outdir + "/" + prefix_out + ".err.log", "w")
            err_file.write(stderr.decode("utf-8"))
            err_file.close()

        # remove the ctl file
        if remove_ctl_file:
            os.remove(work_conf_file)

    return None


def multi_neufile_outlier_removing(
    inpdir,
    generik_conf_file,
    outdir="",
    extention="neu",
    specific_stats=(),
    invert_specific=False,
    remove_ctl_file=True,
):
    if not os.path.exists(inpdir):
        os.makedirs(inpdir)
    os.chdir(inpdir)

    wildcarded_path = inpdir + "/" + "*." + extention

    listofenufile = glob.glob(wildcarded_path)

    if len(listofenufile) == 0:
        log.warning("no files found in specified directory ...")
        log.warning(wildcarded_path)

    listofenufile = keeping_specific_stats(
        listofenufile, specific_stats, invert=invert_specific
    )

    log.warning("removing outliers of %s timeseries", len(listofenufile))

    for enu in sorted(listofenufile):
        neufile_outlier_removing(
            enu, generik_conf_file, outdir=outdir, remove_ctl_file=remove_ctl_file
        )

    return None


def momfile_trend_processing(
    inp_momfile, generik_conf_file, outdir="", remove_ctl_file=True
):
    momfile_inp_name = os.path.basename(inp_momfile)
    inpdir = os.path.dirname(inp_momfile)
    if outdir == "":
        outdir = inpdir
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    work_conf_file = inpdir + "/" + momfile_inp_name.replace(".mom", "") + "_trend.ctl"
    prefix_out_name = momfile_inp_name.replace("_pre.mom", "")
    momfile_out_name = momfile_inp_name.replace("pre.mom", "out.mom")
    momfile_out_path = outdir + "/" + momfile_out_name
    shutil.copyfile(generik_conf_file, work_conf_file)
    utils.replace(work_conf_file, "DataFile", "DataFile            " + momfile_inp_name)
    utils.replace(work_conf_file, "DataDirectory", "DataDirectory       " + inpdir)
    utils.replace(
        work_conf_file, "OutputFile", "OutputFile          " + momfile_out_path
    )

    p = subprocess.Popen(
        "",
        executable="/bin/bash",
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    command = "estimatetrend " + work_conf_file
    log.info("LAUNCHING : %s", command)
    stdout, stderr = p.communicate(command.encode())
    # logs files
    sumfile = outdir + "/" + prefix_out_name + ".sum"
    std_file = open(sumfile, "w")
    std_file.write(stdout.decode("utf-8"))
    std_file.close()
    if stderr:
        log.warning("err.log is not empty, must be checked !")
        log.warning(stderr.decode("utf-8"))
        log.warning("")
        err_file = open(outdir + "/" + prefix_out_name + ".err.log", "w")
        err_file.write(stderr.decode("utf-8"))
        err_file.close()

    # plot a graph
    # matplotlib.use('Agg')
    fig = plt.figure()
    try:
        ax = fig.gca()
        ax.set_xlabel("Date")
        ax.set_ylabel("Displacement (m)")
        data = np.loadtxt(momfile_out_path)
        ax.plot(MJD2dt(data[:, 0]), data[:, 1], "b+")
        ax.plot(MJD2dt(data[:, 0]), data[:, 2], "g.")

        # add offsets
        mom_obj = open(inp_momfile, "r+")
        offset_stk = []
        for line in mom_obj:
            if "offset" in line:
                offset_stk.append(float(line.split()[-1]))
        log.info("offset detected for plot : %s", offset_stk)
        for off in offset_stk:
            plt.axvline(MJD2dt(off), color="r")
        mom_obj.close()

        # add trend & bias
        mom_obj = open(sumfile, "r+")
        for line in mom_obj:
            try:
                if "bias :" in line[0:7]:
                    biasline = line
                if "trend:" in line[0:7]:
                    trendline = line
            except:
                continue
        mom_obj.close()

        #    .text(0.1, 0.15, biasline , horizontalalignment='left',verticalalignment='center',transform=ax.transAxes)
        plt.figtext(
            0.1,
            0.05,
            trendline + biasline,
            horizontalalignment="left",
            verticalalignment="center",
            transform=ax.transAxes,
        )
        plt.suptitle(prefix_out_name)

        # save
        fig.set_size_inches(16.53, 11.69 * 0.6667)
        fig.autofmt_xdate()
        fig.savefig(outdir + "/" + prefix_out_name + ".plt.pdf")
        fig.savefig(outdir + "/" + prefix_out_name + ".plt.png")

    except:
        plt.close(fig)

    # remove the ctl file
    if remove_ctl_file:
        os.remove(work_conf_file)

    return None


def MJD2dt(mjd_in):
    # cf http://en.wikipedia.org/wiki/Julian_day
    try:
        return [dt.datetime(1858, 11, 17) + dt.timedelta(m) for m in mjd_in]
    except:
        return dt.datetime(1858, 11, 17) + dt.timedelta(mjd_in)


def multi_momfile_trend_processing(
    inpdir,
    generik_conf_file,
    outdir="",
    extention="pre.mom",
    remove_ctl_file=True,
    specific_stats=(),
    invert_specific=False,
):
    start = time.time()
    if not os.path.exists(inpdir):
        os.makedirs(inpdir)
    os.chdir(inpdir)
    listofmomfile = glob.glob(inpdir + "/" + "*" + extention)
    listofmomfile = keeping_specific_stats(
        listofmomfile, specific_stats, invert=invert_specific
    )

    log.info("processing % files", len(listofmomfile))

    for imom, mom in enumerate(sorted(listofmomfile)):
        log.info("processing file %s/%s", imom + 1, len(listofmomfile))
        momfile_trend_processing(
            mom, generik_conf_file, outdir=outdir, remove_ctl_file=remove_ctl_file
        )

    log.info("multi_momfile_trend_processing exec time : %s", time.time() - start)
    return None


# ===== FCTS 4 multi_sumfiles_trend_extract =====


def sumfiles_to_statdico(inpdir, specific_stats=(), invert_specific=False):
    """
    this fct search for every sum file in a folder
    a stat dico contains no data
    only the paths to the E,N,U sum files
    statdico[stat] = [path/E.sum,path/N.sum,path/U.sum ]
    for each stat getting the 3 ENU sum files

    Thoses lists will be send in sumfiles_trend_extract
    """
    if not os.path.exists(inpdir):
        os.makedirs(inpdir)
    os.chdir(inpdir)
    listofsumfile = glob.glob(inpdir + "/" + "*" + ".sum")
    listofsumfile = keeping_specific_stats(
        listofsumfile, specific_stats, invert=invert_specific
    )
    listofstat = [
        os.path.basename(sumfil).split(".")[0].split("_")[-2]
        for sumfil in listofsumfile
    ]
    statdico = {}
    for stat in listofstat:
        statdico[stat] = []
        for sumfil in listofsumfile:
            if stat in sumfil:
                statdico[stat].append(sumfil)
    return statdico


def sumfiles_trend_extract(listof3ENUsumfile):
    V, sV = {}, {}
    if len(listof3ENUsumfile) != 3:
        raise Exception("listofENUsumfile != 3")
    for fil in listof3ENUsumfile:
        bnfil = os.path.basename(fil)
        coord = bnfil.split(".")[0][-1]
        for l in open(fil):
            if "trend:" in l:
                f = l.split()
                V[coord] = float(f[1])
                sV[coord] = float(f[3])

    statname = bnfil.split(".")[0].split("_")[-2]
    return statname, V, sV


def get_FLH_from_NEUfile(neufilepath):
    for l in open(neufilepath):
        f = l.split()
        if "Longitude" in l:
            lon = float(f[-1])
        if "Latitude" in l:
            lat = float(f[-1])
        if "Height" in l:
            hau = float(f[-1])
    return lat, lon, hau


def velfile_from_a_list_of_statVsV_tuple(
    listoftup, out_dir, out_prefix, raw_neu_dir="", style="epc"
):
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    """
    make a GLOBK style .vel file
    """

    ### HEADER DEFINITION
    if style == "globk":
        outfile = open(out_dir + "/" + out_prefix + ".vel", "w+")
        outfile.write(
            "* Velocity field created with Hector Runner / p. Sakic - La Rochelle Univ. (FRA) \n"
        )
        outfile.write(
            "*  Long         Lat        Evel    Nvel    dEv     dNv    E +-    N +-    Rne      Hvel     dHv    H +-  Site\n"
        )
        outfile.write(
            "*  deg          deg       mm/yr   mm/yr   mm/yr   mm/yr   mm/yr   mm/yr            mm/yr   mm/yr   mm/yr\n"
        )
        k = 1000.0
    elif style == "epc":
        outfile = open(out_dir + "/" + out_prefix + ".vel.neu", "w+")
        k = 1
    elif style == "csv_renag":
        outfile = open(out_dir + "/" + out_prefix + ".vel.csv", "w+")
        outfile.write("Station,V_North,V_East,V_Up,sV_North,sV_East,sV_Up\n")
        k = 1000.0
    elif style == "csv_renag_xyz":
        outfile = open(out_dir + "/" + out_prefix + ".vel.csv", "w+")
        outfile.write("Station,V_X,V_Y,V_Z,sV_X,sV_Y,sV_Z\n")
        k = 1000.0
    elif style == "dataframe":
        column_names = [
            "Station",
            "Latitude",
            "Longitude",
            "V_North",
            "V_East",
            "V_Up",
            "sV_North",
            "sV_East",
            "sV_Up",
        ]
        lines_stk = []
        k = 1

    ### FILLING DEFINITION
    for tup in listoftup:
        stat, V, sV = tup
        try:
            neufile = utils.regex2filelist(raw_neu_dir, stat)[0]
            lat, lon, hau = get_FLH_from_NEUfile(neufile)
        except:
            lat, lon, hau = 0, 0, 0
        if style == "globk":
            line = "{:11.5f}{:11.5f} {:8.2f}{:8.2f}{:8.2f}{:8.2f}{:8.2f}{:8.2f} {:6.3f}  {:8.2f}{:8.2f}{:8.2f} {}\n".format(
                lon,
                lat,
                V["E"] * k,
                V["N"] * k,
                0.0,
                0.0,
                sV["E"] * k,
                sV["N"] * k,
                0.0,
                V["U"] * k,
                0.0,
                sV["U"] * k,
                stat + "_GPS",
            )
        elif style == "epc":
            line = "{} {} {} {} {} {} {} {}\n".format(
                stat,
                lat,
                reffram.wrapTo180(lon),
                V["N"] * k,
                V["E"] * k,
                sV["N"] * k,
                sV["E"] * k,
                0,
            )
        elif style == "csv_renag":
            line = "{},{:10.5f},{:10.5f},{:10.5f},{:10.5f},{:10.5f},{:10.5f}\n".format(
                stat,
                V["N"] * k,
                V["E"] * k,
                V["U"] * k,
                sV["N"] * k,
                sV["E"] * k,
                sV["U"] * k,
            )
        elif style == "csv_renag_xyz":
            line = "{},{:10.5f},{:10.5f},{:10.5f},{:10.5f},{:10.5f},{:10.5f}\n".format(
                stat,
                V["X"] * k,
                V["Y"] * k,
                V["Z"] * k,
                sV["X"] * k,
                sV["Y"] * k,
                sV["Z"] * k,
            )
        elif style == "dataframe":
            line = [
                stat,
                lat,
                lon,
                V["N"] * k,
                V["E"] * k,
                V["U"] * k,
                sV["N"] * k,
                sV["E"] * k,
                sV["U"] * k,
            ]
            lines_stk.append(line)

        if style != "dataframe":
            outfile.write(line)

    #### FINALISATION
    if style == "dataframe":
        DF = pd.DataFrame(lines_stk, columns=column_names)
        utils.pickle_saver(DF, out_dir, out_prefix + "_DataFrame")
        return DF
    else:
        outfile.close()
        return None


# def velNEUfile_from_a_list_of_statVsV_tuple(listoftup,out_dir, out_prefix , raw_neu_dir=''):
#    """ make a dirty velocity file compatible with EPC """
#    outfile = open(out_dir +'/' + out_prefix + '.vel.neu','w+')
#    for tup in listoftup:
#        stat , v , sV = tup
#        try:
#            neufile = utils.regex2filelist(raw_neu_dir,stat)[0]
#            lat , lon , hau = get_FLH_from_NEUfile(neufile)
#        except:
#            lat , lon , hau = 0,0,0
#
#        line = '{} {} {} {} {} {} {} {}\n'.format(stat,lat,refframe.wrapTo180(lon),v['N'],v['E'],sV['N'],sV['E'],0)
#        outfile.write(line)
#    return None


def multi_sumfiles_trend_extract(
    inp_dir,
    out_dir,
    out_prefix,
    raw_neu_dir="",
    specific_stats=(),
    invert_specific=False,
    style="epc",
):
    """style = globk OR epc
    make a GLOBK style .vel file or
    make a dirty velocity file compatible with EPC"""

    utils.create_dir(out_dir)

    statdico = sumfiles_to_statdico(
        inp_dir, specific_stats, invert_specific=invert_specific
    )
    listoftup = []
    for stat, listof3ENUsumfile in statdico.items():
        tup = sumfiles_trend_extract(listof3ENUsumfile)
        if tup[0] != stat:
            raise Exception("tup[0] != stat")
        listoftup.append(tup)
    listoftup.sort(key=lambda x: x[0])
    velfile_from_a_list_of_statVsV_tuple(
        listoftup, out_dir, out_prefix, raw_neu_dir, style
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 28/02/2025 18:52:44

@author: psakic
"""
import re
import subprocess

import geodezyx.operational.gins_runner.gins_common as gynscmn
import geodezyx.operational.gins_runner.gins_dirs_gen as gynsgen
import geodezyx.operational.gins_runner.gins_dirs_run as gynsrun
import geodezyx.operational.gins_runner.gins_bd_update as gynsbdu
import geodezyx.operational.gins_runner.gins_orbclk_concat as gynsorbclkcat
import datetime as dt

from geodezyx import utils, operational

import multiprocessing as mp
import os
import shutil
import glob
import time

from geodezyx import files_rw, conv

spotgins_wrap = None
spotgins_wrap2 = None

#### Import the logger
import logging

log = logging.getLogger("geodezyx")


LAST_VERSION_VALIDE = "VALIDE_25_1"
#DIR_SPOTGINS_DEFAULT = ""DIR_SPOTGINS_G20_GE.yml""
DIR_SPOTGINS_DEFAULT = "DIR_SPOTGINS_G20_GE_VALIDE_25_1.yml"

def spotgins_run(
    rnxs_path_inp,
    results_folder_inp,
    nprocs=8,
    version=LAST_VERSION_VALIDE,
    const="GE",
    director_generik_path_inp=None,
    director_name_prefix_inp="",
    stations_file_inp=None,
    oceanload_file_inp=None,
    options_prairie_file_inp=None,
    stations_master_file_inp=None,
    force=False,
    no_archive=False,
    no_clean_tmp=False,
    no_updatebd=False,
    no_concat_orb_clk=False,
    verbose=False,
    updatebd_login="",
):
    """
    Run the SPOTGINS process on the provided RINEX paths using multiprocessing.

    Parameters
    ----------
    rnxs_path_inp : list
        List of input RINEX file paths.
    results_folder_inp : str
        Path to the archive folder.
    nprocs : int, optional
        Number of processes to use for multiprocessing. Default is 8.
    version : str, optional
        Version of the GINS software to use.
         Default is the global variable LAST_VERSION_VALIDE.
    const : str, optional
        GNSS constellation to use. Default is "G".
    director_generik_path_inp : str, optional
        Path to the generic director file. Default is None.
    director_name_prefix_inp : str, optional
        Prefix for the director name. Default is an empty string.
    stations_file_inp : str, optional
        Path to the stations file. Default is None.
    oceanload_file_inp : str, optional
        Path to the ocean load file. Default is None.
    options_prairie_file_inp : str, optional
        Path to the options prairie file. Default is None.
    stations_master_file_inp : str, optional
        Path to the stations master file. Default is None.
    force : bool, optional
        Force the generation and run of directors even if they already exist. Default is False.
    no_archive : bool, optional
        If True, do not archive the results. Default is False.
    no_clean_tmp : bool, optional
        If True, do not clean the temporary files. Default is False.
    no_updatebd : bool, optional
        If True, do not update the BDGINS repository. Default is False.
    no_concat_orb_clk : bool, optional
        If True, do not concatenate the orbit and clock files prior to the main processing.
         Default is False.
    verbose : bool, optional
        If True, enable verbose logging. Default is False.
    updatebd_login : str
        The login to connect to the remote `tite` GINS server to update the database.
        We assume that SSH keys have been exchanged to automatize the connexion
    Returns
    -------
    None
    """

    ##### sort the input list
    rnxs_path_use = utils.listify(rnxs_path_inp)
    rnxs_path_use = operational.rinex_table_from_list(
        rnxs_path_use, site9_col=True, size_col=False
    )
    rnxs_path_use = rnxs_path_use["path"].to_list()

    if len(rnxs_path_use) == 0:
        log.error("No input RINEX files to process, skip")
        return

    ##### get the paths of the files needed for SPOTGINS
    dirgen_use, stfi_use, oclo_use, opra_use, siteid9_use = get_spotgins_files(
        director_generik_path_inp,
        stations_file_inp,
        oceanload_file_inp,
        options_prairie_file_inp,
        stations_master_file_inp,
    )

    ##### Need timespan
    if not no_updatebd or not no_concat_orb_clk:
        rnxs_dates = [conv.rinexname2dt(e) for e in rnxs_path_use]
        rnxs_dates = [e for e in rnxs_dates if not e is None]
        date_min, date_max = (
            min(rnxs_dates) - dt.timedelta(days=1),
            max(rnxs_dates) + dt.timedelta(days=1),
        )
    else:
        date_min = dt.datetime(2000, 5, 3)
        date_max = None
    ##### Update the database ################
    if not no_updatebd:
        gynsbdu.bdgins_update(
            date_srt=date_min, date_end=date_max, dir_bdgins="", login=updatebd_login
        )

    ##### concatenate hor/orb ################
    if not no_concat_orb_clk:
        gynsorbclkcat.concat_orb_clk(date_min, date_max, nprocs=nprocs)

    ##### Multi-processing Wrapper ################
    global spotgins_wrap

    def spotgins_wrap(rnx_mono_path_inp):
        ######## QUICK ARCHIVING CHECK ############# # Check if the solution is already archived
        if not force and check_arch_sol(
            rnx_mono_path_inp, results_folder_inp, verbose=verbose
        ):
            log.info(f"Solution already archived for {rnx_mono_path_inp}, skip")
            return None

        ######## DIRECTORS GENERATION ########
        dirr, tmp_folder = gynsgen.gen_dirs_rnxs(
            rnx_paths_inp=rnx_mono_path_inp,
            director_generik_path=dirgen_use,
            director_name_prefix="SPOTGINS",
            stations_file=stfi_use,
            oceanload_file=oclo_use,
            auto_stations_file=False,
            auto_oceanload=False,
            options_prairie_file=opra_use,
            auto_interval=False,
            force=force,
            verbose=verbose,
            sites_id9=siteid9_use,
        )

        ######## DIRECTORS RUN ###############
        const_use = const_adapt(const, dirr, verbose=verbose)
        # const_use = "GE" # ASG use always GE
        opt_gins_90_use = "-const " + const_use
        try:
            gynsrun.run_directors(
                dirr,
                opts_gins_90=opt_gins_90_use,
                version=version,
                cmd_mode="exe_gins_dir",
                force=force,
                verbose=verbose,
                sleep_time_max=nprocs * 10**-1,  # eq 128 procs -> 12.8s max sleep time
            )
        except Exception as e:
            log.error(f"run_directors error: {e}")
            rm_prov_listing(dirr)

        ######## ARCHIVING ####################
        if not no_archive:
            archiv_gins_run(dirr, results_folder_inp, verbose=verbose)

        ######## CLEANING ####################
        if not no_clean_tmp and os.path.exists(tmp_folder):
            if verbose:
                log.info(f"Cleaning temporary folder {tmp_folder}")
            shutil.rmtree(tmp_folder)
        return None

    ##### END Multi-processing Wrapper END ################

    global spotgins_wrap2

    def spotgins_wrap2(*args):
        try:
            return spotgins_wrap(*args)
        except Exception as e:
            log.error(f"spotgins_wrap error: {e}")
            return None  # or something that won't break your pipeline

    pool = mp.Pool(processes=nprocs)
    try:
        res_raw = pool.map(spotgins_wrap2, rnxs_path_use, chunksize=1)
    except Exception as e:
        log.error("pool.map error: %s", e)
    pool.close()

    nrnxs = len(rnxs_path_use)
    log.info("multi run end, all of the %i RINEXs have been procssed", nrnxs)

    return


def get_spotgins_files(
    director_generik_path_inp,
    stations_file_inp,
    oceanload_file_inp,
    options_prairie_file_inp,
    stations_master_file_inp,
):
    """
    Retrieve the paths of the files needed for the SPOTGINS process.

    Parameters
    ----------
    director_generik_path_inp : str
        Path to the generic director file.
    stations_file_inp : str
        Path to the stations file.
    oceanload_file_inp : str
        Path to the ocean load file.
    options_prairie_file_inp : str
        Path to the options prairie file.
    stations_master_file_inp : str
        Path to the stations master file.

    Returns
    -------
    tuple
        A tuple containing the paths to the director file,
        stations file, ocean load file, options prairie file,
        and site IDs.
    """
    if not gynscmn.get_spotgins_path() and (
        not stations_file_inp or not oceanload_file_inp or not options_prairie_file_inp
    ):
        raise ValueError(
            "SPOTGINS path not set ($SPOTGINS_DIR) and stations_file/oceanload_file/options_prairie_file not provided"
        )
    else:
        sptgns_path = gynscmn.get_spotgins_path()

    if director_generik_path_inp:
        dirgen_use = director_generik_path_inp
    else:
        dirgen_use = os.path.join(
            sptgns_path, "metadata", "directeur", DIR_SPOTGINS_DEFAULT
        )

    if stations_file_inp:
        stfi_use = stations_file_inp
    else:
        stfi_use = os.path.join(sptgns_path, "metadata", "stations", "station_file.dat")

    if oceanload_file_inp:
        oclo_use = oceanload_file_inp
    else:
        oclo_use = os.path.join(
            sptgns_path, "metadata", "oceanloading", "load_fes2014b_cf.spotgins"
        )

    if options_prairie_file_inp:
        opra_use = options_prairie_file_inp
    else:
        opra_use = os.path.join(
            sptgns_path, "metadata", "directeur", "options_prairie_static"
        )

    if stations_master_file_inp and os.path.isfile(stations_master_file_inp):
        siteid9_use = files_rw.read_spotgins_masterfile(stations_master_file_inp)
        siteid9_use = siteid9_use["NAME"]
    elif sptgns_path:
        stations_master_file = os.path.join(
            sptgns_path, "metadata", "stations", "station_master_file.dat"
        )
        siteid9_use = files_rw.read_spotgins_masterfile(stations_master_file)
        siteid9_use = siteid9_use["NAME"]
    else:
        siteid9_use = None

    return dirgen_use, stfi_use, oclo_use, opra_use, siteid9_use


def rm_prov_listing(dir_inp):
    """
    Remove potential PROV folders and related files in the listing folder.

    If a processing is interrupted this temp PROV folder remains,
    occupy a lot of space and can eventually crashs the whole machine/container.

    Thus it must be removed ASAP


    Parameters
    ----------
    dir_inp : str
        Path to the input directeur whose associated PROV folders and files
        need to be removed.

    Functionality
    -------------
    1. Identifies and removes all PROV folders associated with the given directeur.
    2. Identifies and removes specific files (e.g., `.qsub` and `.sh` files)
       related to the directory in the batch listing folder.

    Returns
    -------
    None
    """
    # Get the path to the batch listing folder
    li_batch_fld = os.path.join(gynscmn.get_gin_path(True), "batch", "listing")
    # Extract the base name of the input directory
    dir_basename = os.path.basename(dir_inp)

    # Find and remove all PROV folders associated with the directory
    prov_lis = glob.glob(os.path.join(li_batch_fld, "PROV" + "*" + dir_basename + "*"))
    for prov in prov_lis:
        log.warning("removing PROV folder %s", prov)
        shutil.rmtree(prov)

    # Find and remove specific files (e.g., `.qsub` and `.sh`) in the batch listing folder
    lifil_lis = glob.glob(os.path.join(li_batch_fld, dir_basename) + "*")
    for f in lifil_lis:
        if f.endswith(".qsub") or f.endswith("sh"):
            os.remove(f)

    return None


def archiv_gins_run(dir_inp, archive_folder, verbose=True):
    """
    Archive the GINS run results to the specified archive folder.

    Parameters
    ----------
    dir_inp : str
        Path to the directory containing the GINS run results.
    archive_folder : str
        Path to the archive folder.
    verbose : bool, optional
        If True, enable verbose logging. Default is True.

    Returns
    -------
    None
    """

    if not dir_inp:
        return None

    time.sleep(1)  # wait a proper GINS end

    # do first the PROV folder cleaning
    rm_prov_listing(dir_inp)

    dir_basename = os.path.basename(dir_inp)
    site_id9 = re.search(r"_(....00\w{3})_", dir_inp).group(1)
    arch_fld_site = str(os.path.join(archive_folder, site_id9))
    if not os.path.exists(arch_fld_site):
        os.makedirs(arch_fld_site)

    # get input directories
    dir_batch_fld = os.path.join(gynscmn.get_gin_path(True), "data", "directeur")
    li_batch_fld = os.path.join(gynscmn.get_gin_path(True), "batch", "listing")
    sol_batch_fld = os.path.join(gynscmn.get_gin_path(True), "batch", "solution")
    stat_batch_fld = os.path.join(gynscmn.get_gin_path(True), "batch", "statistiques")

    # get destination
    dir_arch_fld = os.path.join(arch_fld_site, "010_directeurs")
    li_arch_fld = os.path.join(arch_fld_site, "020_listings")
    sol_arch_fld = os.path.join(arch_fld_site, "030_solutions")
    stat_arch_fld = os.path.join(arch_fld_site, "040_statistiques")
    trsh_arch_fld = os.path.join(arch_fld_site, "090_trash")

    # create directories
    for arch_fld in [
        dir_arch_fld,
        li_arch_fld,
        sol_arch_fld,
        stat_arch_fld,
        trsh_arch_fld,
    ]:
        if not os.path.exists(arch_fld):
            os.makedirs(arch_fld)

    log.info(f"archive {dir_basename} run in {arch_fld_site}")
    # move files to their destination
    for batch_fld, arch_fld in zip(
        [dir_batch_fld, li_batch_fld, sol_batch_fld, stat_batch_fld],
        [dir_arch_fld, li_arch_fld, sol_arch_fld, stat_arch_fld],
    ):

        for f in glob.glob(os.path.join(batch_fld, dir_basename) + "*"):
            if f.endswith(".qsub") or f.endswith("sh"):
                os.remove(f)
                continue

                # if verbose:
                #    log.info(f"skip archive of qsub/sh file {f}")
                # continue

            if verbose:
                log.info(f"Archiving {f} to {arch_fld}")

            # check if the file is already in the archive (crash if yes)
            # and move it to the trash folder
            f_bn = os.path.basename(f)
            f_prex = os.path.join(arch_fld, f_bn)
            if os.path.exists(f_prex):
                timstp = utils.get_timestamp()
                f_prex_trsh = os.path.join(trsh_arch_fld, f_bn + "_" + timstp)
                shutil.move(f_prex, f_prex_trsh)
            # move the actual correct file
            shutil.move(f, arch_fld)

    # compress the listings
    for f in glob.glob(os.path.join(li_arch_fld, dir_basename) + "*"):
        if not f.endswith("gz"):
            if verbose:
                log.info(f"Compressing listing {os.path.basename(f)}")
            utils.gzip_compress(f, rm_inp=True)
    return None


def check_arch_sol(rnx_path_inp, archive_folder_inp, verbose=True):
    """
    Check if the solution for the given RINEX file is already archived.

    Parameters
    ----------
    rnx_path_inp : str
        Path to the input RINEX file.
    archive_folder_inp : str
        Path to the archive folder.
    verbose : bool, optional
        If True, enable verbose logging. Default is True.

    Returns
    -------
    bool
        True if the solution is found in the archive, False otherwise.
    """
    epo = conv.rinexname2dt(rnx_path_inp)
    epo_str = epo.strftime("%Y_%j")
    site_id4 = str(os.path.basename(rnx_path_inp)[0:4].upper())

    potential_sol = glob.glob(
        os.path.join(
            archive_folder_inp,
            "*" + site_id4 + "*",
            "030_solutions",
            "*" + site_id4 + "*" + epo_str + "*",
        )
    )

    if len(potential_sol) > 0:
        if verbose:
            log.info(f"Solution(s) found: {potential_sol}")
        return True
    else:
        return False


def const_adapt(const_inp, dir_inp, verbose=True):
    """
    manage
    ERREUR : Galileo integer ambiguity fixing not possible before 7th October 2018.
    solution
    GPS-only before this date
    """

    import yaml

    dir_dic = yaml.load(open(dir_inp), Loader=yaml.FullLoader)
    rnx_date = conv.jjul_cnes2dt(dir_dic["date"]["arc_start"][0])
    if "E" in const_inp and rnx_date < dt.datetime(2018, 10, 7):
        if verbose:
            msg = "Galileo not recommended before 2018-10-07, switch for GPS-only"
            log.warning(msg)
        return "G"
    else:
        return const_inp



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 27/08/2025 22:26:15

@author: psakic
"""

import re
import os
import itertools
import numpy as np
import pandas as pd
from io import StringIO
from geodezyx import conv
from geodezyx import utils


def read_snx_trop(snxfile, dataframe_output=True):
    """
    Read troposphere solutions from Troposphere SINEX
    """

    stat, epoc = [], []
    tro, stro, tgn, stgn, tge, stge = [], [], [], [], [], []

    flagtrop = False

    for line in open(snxfile, "r", encoding="ISO-8859-1"):

        if re.compile("TROP/SOLUTION").search(line):
            flagtrop = True
            continue

        if re.compile("-TROP/SOLUTION").search(line):
            flagtrop = False
            continue

        if line[0] != " ":
            continue
        else:
            fields = line.split()

        if flagtrop:

            stat.append(fields[0].upper())
            if not ":" in fields[1]:
                epoc.append(conv.year_decimal2dt(fields[1]))
            else:
                date_elts_lis = fields[1].split(":")
                yy = int(date_elts_lis[0]) + 2000
                doy = int(date_elts_lis[1])
                sec = int(date_elts_lis[2])

                epoc.append(conv.doy2dt(yy, doy, seconds=sec))

            tro.append(float(fields[2]))
            stro.append(float(fields[3]))
            tgn.append(float(fields[4]))
            stgn.append(float(fields[5]))
            tge.append(float(fields[6]))
            stge.append(float(fields[7]))

    outtuple = list(zip(*sorted(zip(stat, epoc, tro, stro, tgn, stgn, tge, stge))))

    if dataframe_output:
        return tropsinex_data_frame(outtuple)
    else:
        return outtuple


def tropsinex_data_frame(read_sinex_result):
    df_sinex = pd.DataFrame.from_records(list(read_sinex_result)).transpose()
    colnam = ["STAT", "epoc", "tro", "stro", "tgn", "stgn", "tge", "stge"]
    df_sinex.columns = colnam
    cols_numeric = ["tro", "stro", "tgn", "stgn", "tge", "stge"]
    df_sinex[cols_numeric] = df_sinex[cols_numeric].apply(
        pd.to_numeric, errors="coerce"
    )

    return df_sinex


def read_sinex(sinex_path_in, keep_sites_as_index=False, drop_nan=False):
    """
    Read a *coordinate* SINEX file as DataFrame

    Parameters
    ----------
    sinex_path_in : str
        path of the SINEX.
    keep_sites_as_index : Bool, optional
        use the site names as index. The default is False.
    drop_nan : bool or str, optional
        remove the NaN in the DataFrame.
        False: do not remove them
        'all': remove the row if all values are NaN
        'any':  remove the row if at least one value is NaN (dangerous)
        The default is False.

    Returns
    -------
    DFout : DataFrame
        SINEX DataFrame.

    """
    ## Read the blocs
    df_coor = read_sinex_versatile(
        sinex_path_in, "SOLUTION/ESTIMATE", header_line_idx=None
    )
    df_epoc = read_sinex_versatile(
        sinex_path_in, "SOLUTION/EPOCHS", header_line_idx=None
    )

    ## Rename the Epoch DF
    df_epoc.rename(
        columns={
            0: "STAT",
            1: "pt",
            2: "soln",
            3: "?",
            4: "start",
            5: "end",
            6: "mean",
        },
        inplace=True,
    )
    df_epoc.set_index(["STAT", "pt", "soln"], inplace=True)

    ### Rearange the coords DF
    index = df_coor[[2, 3, 4]].drop_duplicates()

    stk_df_coor2_line = []

    for stat, pt, soln in zip(index[2], index[3], index[4]):
        df2 = df_coor[(df_coor[2] == stat) & (df_coor[3] == pt) & (df_coor[4] == soln)]

        dicttmp = dict()

        epoc = df2[5].unique()[0]
        dicttmp["epoc"] = epoc
        dicttmp["STAT"] = stat
        dicttmp["pt"] = pt
        dicttmp["soln"] = soln

        for typ, comp in itertools.product(("STA", "VEL"), ("X", "Y", "Z")):

            typcomp = typ + comp
            try:
                val = df2[df2[1] == typcomp][8].values[0]
                sig = df2[df2[1] == typcomp][9].values[0]
            except:
                val, sig = np.nan, np.nan

            if typ == "VEL":
                typdic = "v"
            else:
                typdic = ""

            dicttmp[typdic + comp.lower()] = val
            dicttmp["s" + typdic + comp.lower()] = sig

        df_coor2_line = pd.DataFrame(dicttmp, index=[0])
        stk_df_coor2_line.append(df_coor2_line)

    df_coor2 = pd.concat(stk_df_coor2_line)
    df_coor2.set_index(["STAT", "pt", "soln"], inplace=True)

    ### Concatenate both
    DFout = pd.concat((df_epoc, df_coor2), axis=1)

    ### remove NaN (might be more epochs than coords...)
    if drop_nan:
        DFout.dropna(inplace=True, how=drop_nan)

    if not keep_sites_as_index:
        DFout.reset_index(inplace=True)

    return DFout


def read_sinex_legacy(snxfile, dataframe_output=True):
    """
    This function is depreciated !!!!
    """

    stat, soln, epoc, ac = [], [], [], []
    x, y, z, sx, sy, sz = [], [], [], [], [], []
    vx, vy, vz, svx, svy, svz = [], [], [], [], [], []

    start, end = [], []

    flagxyz = False
    flagepochs = False

    for line in open(snxfile, "r", encoding="ISO-8859-1"):

        if re.compile("SOLUTION/ESTIMATE").search(line):
            flagxyz = not flagxyz
            continue

        if re.compile("SOLUTION/EPOCHS").search(line):
            flagepochs = not flagepochs
            continue

        if line[0] != " ":
            continue
        else:
            fields = line.split()

        if flagxyz:

            if fields[1] == "STAX":
                split_sp3_name = os.path.basename(snxfile)
                ac.append(split_sp3_name[0:3])
                stat.append(fields[2].upper())
                soln.append(fields[4])
                if not ":" in fields[5]:
                    epoc.append(conv.year_decimal2dt(fields[5]))
                else:
                    date_elts_lis = fields[5].split(":")
                    yy = int(date_elts_lis[0]) + 2000
                    doy = int(date_elts_lis[1])
                    sec = int(date_elts_lis[2])

                    epoc.append(conv.doy2dt(yy, doy, seconds=sec))

                x.append(float(fields[8]))
                sx.append(float(fields[9]))

            if fields[1] == "STAY":
                y.append(float(fields[8]))
                sy.append(float(fields[9]))

            if fields[1] == "STAZ":
                z.append(float(fields[8]))
                sz.append(float(fields[9]))

            if fields[1] == "VELX":
                vx.append(float(fields[8]))
                svx.append(float(fields[9]))

            if fields[1] == "VELY":
                vy.append(float(fields[8]))
                svy.append(float(fields[9]))

            if fields[1] == "VELZ":
                vz.append(float(fields[8]))
                svz.append(float(fields[9]))

        if flagepochs:
            start_val = conv.datestr_sinex_2_dt(fields[4])
            end_val = conv.datestr_sinex_2_dt(fields[5])
            start.append(start_val)
            end.append(end_val)

    if not vx:
        nanlist = [np.nan] * len(x)
        vx, vy, vz, svx, svy, svz = (
            list(nanlist),
            list(nanlist),
            list(nanlist),
            list(nanlist),
            list(nanlist),
            list(nanlist),
        )

    outtuple = list(
        zip(
            *sorted(
                zip(
                    ac,
                    stat,
                    soln,
                    epoc,
                    x,
                    y,
                    z,
                    sx,
                    sy,
                    sz,
                    vx,
                    vy,
                    vz,
                    svx,
                    svy,
                    svz,
                    start,
                    end,
                )
            )
        )
    )

    if dataframe_output:
        return sinex_data_frame(outtuple)
    else:
        stat, soln, epoc, x, y, z, sx, sy, sz, vx, vy, vz, svx, svy, svz, start, end = (
            outtuple
        )
        print(
            "TIPS : this output can be converted directly to a Pandas DataFrame using sinex_data_frame function"
        )
        return (
            stat,
            soln,
            epoc,
            x,
            y,
            z,
            sx,
            sy,
            sz,
            vx,
            vy,
            vz,
            svx,
            svy,
            svz,
            start,
            end,
        )


def sinex_data_frame(read_sinex_result):
    DF_Sinex = pd.DataFrame.from_records(list(read_sinex_result)).transpose()
    colnam = [
        "AC",
        "STAT",
        "soln",
        "epoc",
        "x",
        "y",
        "z",
        "sx",
        "sy",
        "sz",
        "vx",
        "vy",
        "vz",
        "svx",
        "svy",
        "svz",
        "start",
        "end",
    ]
    DF_Sinex.columns = colnam

    cols_numeric = [
        "x",
        "y",
        "z",
        "sx",
        "sy",
        "sz",
        "vx",
        "vy",
        "vz",
        "svx",
        "svy",
        "svz",
    ]

    DF_Sinex[cols_numeric] = DF_Sinex[cols_numeric].apply(
        pd.to_numeric, errors="coerce"
    )

    return DF_Sinex


def read_sinex_versatile(
    sinex_path_in,
    id_block,
    convert_date_2_dt=True,
    header_line_idx=-1,
    improved_header_detection=True,
    verbose=True,
):
    """
    Read a block from a SINEX and return the data as a DataFrame

    Parameters
    ----------
    sinex_path_in : str
        Description param1

    id_block : str
        Name of the block (without "+" or "-")

    convert_date_2_dt : bool
        Try to convert a SINEX formated date as a python datetime

    header_line_idx : int or None
        If the block header contains several lines, use this line index
        Per default, the last (-1)
        For the first line, use 0
        If no header is properly defined, use None

    improved_header_detection : bool
        Improved header detection.
        Works for most cases but sometime the simple version works better.
        (advenced usage)
        Default is True

    verbose : bool
        print the header and its field size
        Default is True

    Returns
    -------
    df : Pandas DataFrame
        Returned DataFrame
    """

    ### remove the + or - if any
    if id_block in ("+", "-"):
        id_block = id_block[1:]

    id_block_strt = r"\+" + id_block
    id_block_end = r"\-" + id_block

    lines_list = utils.extract_text_between_elements_2(
        sinex_path_in, id_block_strt, id_block_end
    )
    lines_list = lines_list[1:-1]

    if not lines_list:
        print("ERR : read_sinex_versatile : no block found, ", id_block)

    #### Remove commented lines
    lines_list_header = []
    lines_list_ok = []
    header_lines = True
    for i_l, l in enumerate(lines_list):
        if not l[0] in (" ", "\n") and header_lines:
            ## here we store the 1st commented lines i.e. the header
            lines_list_header.append(l)
        elif l[0] in (" ", "\n"):
            ## here we store the data lines (not commented)
            header_lines = False
            lines_list_ok.append(l)
        else:
            ## here we skip the final commented lines (can happend)
            continue

    if len(lines_list_header) > 0 and header_line_idx:
        ### define the header
        header_line = lines_list_header[header_line_idx]

        header_split = header_line.split()
        if not improved_header_detection:
            ### Simple case when the columns are splitted with only a single
            fields_size = [len(e) + 1 for e in header_split]
        else:
            ### Smarter case : we search for the n spaces after the column name
            fields_size = []

            for fld_head_split in header_split:
                if fld_head_split[0] == "*":
                    ## test to manage a * as 1st character
                    ## which can be wrongly interpreted in the regex
                    fld_head_regex = re.compile(r"\*" + fld_head_split[1:] + " *")
                else:
                    fld_head_regex_str = fld_head_split + " *"
                    ### update 202110: now the brackets must be escaped (normal btw...)
                    fld_head_regex_str = fld_head_regex_str.replace("[", r"\[")
                    fld_head_regex_str = fld_head_regex_str.replace("]", r"\]")
                    fld_head_regex_str = fld_head_regex_str.replace("(", r"\(")
                    fld_head_regex_str = fld_head_regex_str.replace(")", r"\)")

                    fld_head_regex = re.compile(fld_head_regex_str)

                fld_head_space = fld_head_regex.search(header_line)

                fields_size.append(len(fld_head_space.group()))
                # print(fld_head_space.group())

                # # weak method (210216) archived for legacy
                # fld_head_regex = re.compile(fld_head_split[1:] + " *") #trick:
                # #1st char is removed, because it can be a *
                # #and then screw the regex. This char is re-added at the end
                # #when the len is stored (the "+1" below)
                # fld_head_space = fld_head_regex.search(header_line)
                # fields_size.append(len(fld_head_space.group()) + 1)
                # ### !!!!! something is weird here !!!!!
                # print(fld_head_space.group())
                # ### and you will see a bug !!!
                # ### PS 210216

        if verbose:
            print("INFO : read_sinex_versatile : Auto detected column names/sizes")
            print("**** Raw header line in the file:")
            print(header_line)
            print("**** Splited header for the DataFrame:")
            print(header_split)
            print("**** Size of the fields")
            print(fields_size)

        ### Add the header in the big string
        lines_str_w_head = header_line + "".join(lines_list_ok)

        ### Read the file
        try:
            df = pd.read_fwf(StringIO(lines_str_w_head), widths=fields_size)
        except pd.errors.EmptyDataError as ee:
            print("ERR: something goes wrong in the header index position")
            print("     try to give its right position manually with header_line_idx")
            raise (ee)

        df.set_axis(header_split, axis=1, inplace=True)

        ### Rename the 1st column (remove the comment marker)
        df.rename(columns={df.columns[0]: df.columns[0][1:]}, inplace=True)

    else:  # no header in the SINEX
        lines_str = "".join(lines_list_ok)
        df = pd.read_csv(StringIO(lines_str), header=None, delim_whitespace=True)

    regex_time = "(([0-9]{2}|[0-9]{4}):[0-9]{3}|[0-9]{7}):[0-9]{5}"
    for col in df.columns:
        if convert_date_2_dt and re.match(regex_time, str(df[col].iloc[0])):
            try:
                df[col] = df[col].apply(lambda x: conv.datestr_sinex_2_dt(x))
            except Exception as e:
                print(
                    "WARN : read_sinex_versatile : convert date string to datetime failed"
                )
                print(e)
                pass

    return df


def read_sinex_bench_antenna(sinex_in):
    f = open(sinex_in, "r")

    line_good_stk = []
    for l in f:
        if l[0] == " ":
            line_good_stk.append(l)

    T = StringIO("".join(line_good_stk))

    header_cols = [
        "TYP",
        "STA_",
        "Occ",
        "Code",
        "PT",
        "T",
        "___x/up___",
        "___y/n____",
        "___z/e____",
        "_Data_Start",
        "_Data_End__",
        "__Antenna_type______",
        "Radome",
        "__S/N__",
    ]
    df_antenna = pd.read_table(
        T, delim_whitespace=True, error_bad_lines=False, names=header_cols
    )

    return df_antenna


# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.operational contains functions to run the
GNSS processing software GINS.

it can be imported directly with:
from geodezyx import operational

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

########## BEGIN IMPORT ##########
#### External modules
import datetime as dt
import glob


import os
import re
import shutil
import subprocess
import time
import yaml

#### geodeZYX modules
from geodezyx import conv
from geodezyx import operational
from geodezyx import utils

#### Import the logger
import logging

log = logging.getLogger("geodezyx")

##########  END IMPORT  ##########


def get_gin_path(extended=False):
    """
    Get the GIN path from the environment variable GS_USER.

    Parameters
    ----------
    extended : bool, optional
        If True, returns the path to the 'gin' directory inside the GIN path.
        Defaults to False.

    Returns
    -------
    str or None
        The GIN path or None if the environment variable GS_USER does not exist.
    """
    try:
        gs_user = os.environ["GS_USER"]
    except:
        log.error("env. var. $GS_USER dont exists !!!")
        return None
        # gs_user =  os.environ['HOME']
    if not extended:
        return gs_user
    else:
        return os.path.join(os.environ["GS_USER"], "gin")


def get_spotgins_path():
    try:
        return os.environ["SPOTGINS_DIR"]
    except:
        log.error("env. var. SPOTGINS_DIR dont exists !!!")
        return None


def check_site_stfl(stat, stationfile):
    """
    Check if a station code is present in a station file.

    Parameters
    ----------
    stat : str
        The station code to search for.
    stationfile : str
        The path to the station file.

    Returns
    -------
    bool
        True if the station code is found in the station file, False otherwise.
    """
    boolout = utils.check_regex(stationfile, stat)
    if not boolout:
        log.warning("%s not in %s", stat, stationfile)
        log.warning("check RINEX header and its MARKER NAME field (4-char. code)")
    return boolout


def chek_domes_oclo(domes, oclofile):
    """
    Check if a DOMES number is present in an OCLO file.

    Parameters
    ----------
    domes : str
        The DOMES number to search for.
    oclofile : str
        The path to the OCLO file.

    Returns
    -------
    bool
        True if the DOMES number is found in the OCLO file, False otherwise.
    """
    boolout = utils.check_regex(oclofile, "^   " + str(domes))
    if not boolout:
        log.error("%s not in %s, %s", domes, oclofile, "be sure of what you are doing!")

    return boolout


def find_domes_in_stfl(stat_code, stationfile):
    """
    Find the DOMES number for a station code in a station file.
    """
    fil = open(stationfile)
    for l in fil:
        if stat_code in l:
            f = l.split()
            return f[0], f[1]
    return "00000", "M000"


def check_if_in_gin(file_path_inp, gins_path=None, verbose=True):
    """
    Check if a file is within the GIN path.

    Parameters
    ----------
    file_path_inp : str
        The path of the file to check.
    gins_path : str, optional
        The GIN path to check against. If not provided, it defaults to the 'gin' directory inside the GIN path.
    verbose : bool
        verbose messages. Default is True.

    Returns
    -------
    bool
        True if the file is within the GIN path, False otherwise.
    """
    if not gins_path:
        gins_path = os.path.join(get_gin_path(), "gin")

    real_path_file = os.path.realpath(file_path_inp)
    real_path_gins = os.path.realpath(gins_path)

    if real_path_gins in real_path_file:
        boolout = True
    else:
        boolout = False

    if not boolout and verbose:
        log.warning("%s not in %s", file_path_inp, gins_path)

    return boolout


def copy_in_gin(file_path_inp, temp_data_folder=None, verbose=True):
    """
    Copy a file to the GIN temporary data folder.

    Parameters
    ----------
    file_path_inp : str
        The path of the file to be copied.
    temp_data_folder : str, optional
        The path to the temporary data folder.
        If not provided, it defaults to 'TEMP_DATA' inside the GIN path.
    verbose : bool
        verbose messages. Default is True.

    Returns
    -------
    str
        The path of the copied file in the temporary data folder.
    """
    if not temp_data_folder:
        temp_data_folder = os.path.join(get_gin_path(), "TEMP_DATA")
    if not os.path.exists(temp_data_folder):
        os.makedirs(temp_data_folder)
    if verbose:
        log.info("copy %s > %s", file_path_inp, temp_data_folder)
    shutil.copy2(file_path_inp, temp_data_folder)

    file_path_out = os.path.join(temp_data_folder, os.path.basename(file_path_inp))

    return file_path_out


def bring_to_gin(file_path_inp, temp_data_folder=None, gins_path=None, verbose=True):
    """
    Ensure a file is within the GIN path, copying it to the temporary data folder if necessary.

    Combo of check_if_in_gin and copy_in_gin functions.

    Parameters
    ----------
    file_path_inp : str
        The path of the file to be checked and possibly copied.
    temp_data_folder : str, optional
        The path to the temporary data folder. If not provided, it defaults to 'TEMP_DATA' inside the GIN path.
    gins_path : str, optional
        The GIN path to check against. If not provided, it defaults to the 'gin' directory inside the GIN path.
    verbose : bool
        verbose messages. Default is True.

    Returns
    -------
    str
        The path of the file within the GIN path.
    """
    if not check_if_in_gin(file_path_inp, gins_path=gins_path, verbose=verbose):
        file_path_out = copy_in_gin(file_path_inp, temp_data_folder=temp_data_folder, verbose=verbose)
    else:
        file_path_out = file_path_inp

    return file_path_out

def make_path_ginsstyle(pathin):
    """
    Convert an absolute path containing '/gin/' to a GINS director-compatible path.

    Parameters
    ----------
    pathin : str
        The input path, which must be an absolute path
        containing '/gin/'.

    Returns
    -------
    str or None
        The converted path with '.temp.gin' as the root directory,
         or None if the input path does not contain '/gin/'.
    """
    if pathin.startswith(".temp.gin"):
        log.info("already a director compatible path: %s", pathin)
        return pathin

    if not "/gin/" in pathin:
        log.error("not /gin/ in %s", pathin)
        return None

    rnx_path_lis = pathin.split("/")
    i_gin = rnx_path_lis.index("gin")
    rnx_path_lis2 = rnx_path_lis[i_gin:]
    rnx_path_lis2[0] = ".temp.gin"
    pathout = os.path.join(*rnx_path_lis2)
    return pathout

def rm_old_tmp_shel(directory='/tmp', age_sec=600):
    """
    Removes files in the specified directory older than the given age in minutes.

    Parameters:
    directory (str): The directory to search in.
    age_sec (int): Age of files in minutes to delete. Default is 10 minutes.
    """
    try:
        # Call the `find` command to delete files older than the specified age
        subprocess.run(
            ['find', directory, '-type', 'f', '-mmin', f'+{age_sec}', '-delete'],
            check=True
        )
    except subprocess.CalledProcessError as e:
        log.error(f"Error: {e.stderr}")


def get_temp_data_gins_path():
    """
    Get the path to the temporary data folder within the GIN directory.
    """

    gins_path = get_gin_path()

    # be sure there is a TEMP DATA folder
    temp_data_folder = os.path.join(gins_path, "gin", "TMP_GYNS")

    if not os.path.exists(temp_data_folder):
        os.makedirs(temp_data_folder)

    return temp_data_folder

def check_solution(dir_name_inp, gin_path_inp=None, verbose=True):
    """
    Check if a solution already exists for the given directory name.
    """

    dir_name_inp = os.path.basename(dir_name_inp)

    if not gin_path_inp:
        gin_path = get_gin_path()
    else:
        gin_path = gin_path_inp
    # check if the solution already exists
    sol_folder = os.path.join(gin_path, "gin", "batch", "solution")
    sols_matching = glob.glob(sol_folder + "/" + dir_name_inp + "*")
    if len(sols_matching) > 0 and verbose:
        log.info("Solution %s exists for %s", sols_matching, dir_name_inp)
    return sols_matching





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 26/02/2025 11:07:24

@author: psakic
"""

#### Import the logger
import os
import time
import datetime as dt
import subprocess
import numpy as np

#### Import geodezyx GINS submodules
import geodezyx.operational.gins_runner.gins_common as gynscmn

#### geodeZYX modules
from geodezyx import utils

#### Import the logger
import logging

log = logging.getLogger("geodezyx")


def run_directors(
    dir_paths_inp,
    opts_gins_pc="",
    opts_gins_90="",
    version="OPERA",
    cmd_mode="exe_gins_dir",
    force=False,
    verbose=True,
    sleep_time_max=10,
):
    """
    Executes GINS commands for a single or multiple directories, handling different modes and retry logic.

    Parameters
    ----------
    dir_paths_inp : str or list
        Path(s) to the directory/directories to process. Can be a single directory (str) or a list of directories (list).
    opts_gins_pc : str, optional
        Options for the GINS PC command. Default is an empty string.
    opts_gins_90 : str, optional
        Options for the GINS 90 command. Default is an empty string.
    version : str, optional
        Version of the GINS software to use. Default is "OPERA".
    cmd_mode : str, optional
        Mode of execution. Can be "ginspc", "exe_gins_dir", or "exe_gins_fic". Default is "exe_gins_dir".
    force : bool, optional
        If True, forces execution even if solutions already exist. Default is False.
    verbose : bool, optional
        If True, enables verbose logging. Default is True.
    sleep_time_max : int, optional
        Maximum sleep time (in seconds) to avoid conflicts during parallel execution. Default is 10.

    Returns
    -------
    None
        The function does not return any value.

    Notes
    -----
    - Handles both single and multi-directory modes.
    - Automatically detects `.fic` files and adjusts the command mode accordingly.
    - Implements retry logic for failed executions.
    - Logs execution details, including start and end times, and handles parallel execution conflicts.
    """
    # Multi or Single Mode ?
    if type(dir_paths_inp) is list:
        multimode = True
        director_path_lis = dir_paths_inp
        log.info("******** DIRECTORS RUNS *******")
    elif type(dir_paths_inp) is str:
        multimode = False
        director_path_lis = [dir_paths_inp]
    else:
        log.error("check the rinex_paths_in !!!")
        return None

    director_path_lis = list(sorted(director_path_lis))

    ndir = len(director_path_lis)

    for i, dir_path in enumerate(director_path_lis):
        dir_nam = os.path.basename(dir_path)

        if multimode:
            log.info(f"======== {i + 1} / {ndir} ======== ")
        log.info("start %s at %s", dir_nam, dt.datetime.now())
        if dir_path[-4:] == ".fic":
            cmd_mode = "exe_gins_fic"
            log.debug("input file ends with .fic, fic_mode is activated")
        start = time.time()

        if dir_path[-1] == "~":
            log.warning("geany ~ temp file, skiping this dir. ")
            continue

        sols_exist = gynscmn.check_solution(os.path.basename(dir_path), verbose=verbose)
        if len(sols_exist) > 0 and not force:
            log.info("solution %s already exists, skipping ...", sols_exist)
            continue

        ogpc = "-F" + opts_gins_pc
        og90 = opts_gins_90
        vers = version

        # log.info("options ginsPC: %s / gins90: %s", ogpc, og90)

        if "IPPP" in og90 and cmd_mode != "ginspc":
            _check_dir_keys(dir_path)

        ### cmd must be a string here... (and not a list of small str...)
        if cmd_mode == "ginspc":
            cmd = " ".join(("ginspc.bash", ogpc, dir_nam, og90, "-v", vers, "-f"))
        elif cmd_mode == "exe_gins_fic":
            cmd = " ".join(("exe_gins", "-fic", dir_nam, "-v", vers, og90))
        elif cmd_mode == "exe_gins_dir":
            cmd = " ".join(("exe_gins", "-dir", dir_nam, "-v", vers, og90))
        else:
            log.error("mode not recognized !!!")
            return None

        gins_path = gynscmn.get_gin_path()
        log_path = utils.create_dir(os.path.join(gins_path, "python_logs"))
        log_path = os.path.join(gins_path, "python_logs", dir_nam + ".log")

        ## artifical sleep to avoid conflict when parallelizing
        if sleep_time_max > 0:
            time.sleep(np.random.randint(1, sleep_time_max * 1000) * 10**-3)

        ## parameters for retry
        itry = 0
        itry_max = 2
        retry_str = ""
        timeout = 300

        while itry <= itry_max:
            itry += 1

            if verbose:
                log.info("submit. command : %s", cmd)

            process = subprocess.run(
                [cmd],
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                executable="/bin/bash",
                timeout=timeout
            )

            # gynscmn.check_gins_exe(open(log_path), director_name)
            #
            # with open(log_path + ".exec", "w") as f:
            #     f.write("exec time : " + str(time.time() - start))
            #
            ok_sol = gynscmn.check_solution(dir_nam, verbose=verbose)
            if ok_sol:
                log.info("solution ok for: %s :) %s", dir_nam, retry_str)
            else:
                log.error("no solution for: %s :( %s", dir_nam, retry_str)
                #log.error("STDOUT: %s", process.stdout)
                #log.error("STDERR: %s", process.stderr)

            if not ok_sol and check_for_retry(process.stderr) and itry < itry_max:
                log.warning(f"retryable directeur {dir_nam} (try {itry} / {itry_max})")
                itry = itry + 1
                retry_str = f"(try {itry} / {itry_max})"
                time.sleep(3)
            else:
                itry = itry_max + 1

        log.info(
            "run %s end at %s (exec: %8.3f s)",
            dir_nam,
            dt.datetime.now(),
            time.time() - start,
        )

    return None


def check_for_retry(stderr_inp):
    if "exe_gins_recup_fic.sh: 67: [: =: unexpected operator" in stderr_inp:
        return True
    else:
        return False

def _check_dir_keys(director_path_inp):
    for grepstr in (
        "userext_gps__qualiteorb",
        "userext_gps__haute_freq",
        "userext_gps__hor_interp",
        "GPS__QUALITEORB",
        "GPS__HAUTE_FREQ",
        "GPS__HOR_INTERP",
    ):
        grep_out = utils.grep(director_path_inp, grepstr)
        if grep_out == "":
            log.warning("IPPP mode on, but no %sin the dir !!!", grepstr)


#######################################################################################################

#  ______                _   _                                                             _
# |  ____|              | | (_)                                                           | |
# | |__ _   _ _ __   ___| |_ _  ___  _ __     __ _ _ __ __ ___   _____ _   _  __ _ _ __ __| |
# |  __| | | | '_ \ / __| __| |/ _ \| '_ \   / _` | '__/ _` \ \ / / _ \ | | |/ _` | '__/ _` |
# | |  | |_| | | | | (__| |_| | (_) | | | | | (_| | | | (_| |\ v /  __/ |_| | (_| | | | (_| |
# |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|  \__, |_|  \__,_| \_/ \___|\__, |\__,_|_|  \__,_|
#                                             __/ |                     __/ |
#                                            |___/                     |___/



########## OLD FUNCTIONS ##########





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 10/08/2025 11:38:34

@author: psakic
"""

from geodezyx import utils, conv
import geodezyx.operational.gins_runner.gins_common as gynscmn
import os
import subprocess
import datetime as dt
import multiprocessing as mp
import logging
# Import the logger
log = logging.getLogger("geodezyx")

def concat_orb_clk(date_srt, date_end, nprocs=1, prod="G20", verbose=True):
    """
    Concatenate orbit and clock files for a given date range using multiprocessing.

    Parameters
    ----------
    date_srt : datetime
        Start date for the concatenation process.
    date_end : datetime
        End date for the concatenation process.
    nprocs : int
        Number of processes to use for multiprocessing.
    prod : str, optional
        Product type (e.g., "G20"). Default is "G20".
    verbose : bool, optional
        If True, enable verbose logging. Default is True.

    Returns
    -------
    None
    """

    def _chk_cat_orbclk(orbclk_out, silent=False):
        """
        Check if the concatenated orbit/clock file exists.

        Parameters
        ----------
        orbclk_out : str
            Path to the output orbit/clock file.
        silent : bool, optional
            If True, suppress logging. Default is False.

        Returns
        -------
        bool
            True if the file exists, False otherwise.
        """
        orbclk_out = orbclk_out + ".gz"
        if os.path.isfile(orbclk_out):
            log.info("%s here :)", orbclk_out)
            return True
        elif not silent:
            log.error("%s not here :(", orbclk_out)
            return False
        else:
            return False

    def _run_cat_orbclk(cmd, out_fil):
        """
        Run the command to generate the concatenated orbit/clock file.

        Parameters
        ----------
        cmd : str
            Command to execute.
        out_fil : str
            Path to the output file.

        Returns
        -------
        None
        """
        if _chk_cat_orbclk(out_fil, silent=True):
            return None

        if verbose:
            log.info(cmd)
        subprocess.run(
            cmd,
            executable="/bin/bash",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
        )
        utils.gzip_compress(out_fil, rm_inp=True)
        _chk_cat_orbclk(out_fil)
        return None

    global cat_orbclk_wrap

    def cat_orbclk_wrap(date_inp):
        """
        Wrapper function to process a single date for orbit/clock concatenation.

        Parameters
        ----------
        date_inp : datetime
            Date to process.

        Returns
        -------
        tuple
            Paths to the concatenated orbit and clock files.
        """
        jjul_srt = str(conv.dt2jjul_cnes(date_inp - dt.timedelta(days=0)))
        jjul_end = str(conv.dt2jjul_cnes(date_inp + dt.timedelta(days=2)))
        gs_user = gynscmn.get_gin_path(extended=False)
        gins_data = gynscmn.get_gin_path(extended=True) + "/data"

        orb_out = os.path.join(
            gs_user, "GPSDATA", "_".join((prod + "ORB", "AUTOM", jjul_srt, jjul_end))
        )
        cmd_orb = " ".join(
            ["rapat_orb_gnss.sh", jjul_srt, jjul_end, "3", gins_data, orb_out, "0"]
        )

        clk_out = os.path.join(
            gs_user, "GPSDATA", "_".join((prod, "AUTOM", jjul_srt, jjul_end))
        )
        cmd_clk = " ".join(["get_hor_hautes", jjul_srt, jjul_end, prod, clk_out])

        for cmd, out_fil in [(cmd_orb, orb_out), (cmd_clk, clk_out)]:
            _run_cat_orbclk(cmd, out_fil)

        return orb_out, clk_out

    # Generate a list of dates to process
    date_lis = conv.dt_range(date_srt, date_end)

    # Use multiprocessing to process the dates
    pool = mp.Pool(processes=nprocs)
    try:
        _ = pool.map(cat_orbclk_wrap, date_lis, chunksize=1)
    except Exception as e:
        log.error("error in the pool.map : %s", e)
    pool.close()

    return None


def main():
    """
    Command-line interface for concatenating orbit and clock files.
    """
    import argparse
    from datetime import datetime

    parser = argparse.ArgumentParser(
        description="Concatenate orbit and clock files for a given date range using multiprocessing."
    )
    parser.add_argument('-s', '--date_srt', type=str, required=True,
                        help='Start date for the concatenation process (format: YYYY-MM-DD)')
    parser.add_argument('-e', '--date_end', type=str, required=True,
                        help='End date for the concatenation process (format: YYYY-MM-DD)')
    parser.add_argument('-n', '--nprocs', type=int, default=1,
                        help='Number of processes to use for multiprocessing (optional). Default is 1')
    parser.add_argument('-p', '--prod', type=str, default="G20",
                        help='Product type (optional). Default is "G20"')
    parser.add_argument('-v', '--verbose', action='store_true', default=True,
                        help='Enable verbose logging (optional). Default is True')

    args = parser.parse_args()

    # Convert string dates to datetime objects
    try:
        date_srt = datetime.strptime(args.date_srt, '%Y-%m-%d')
        date_end = datetime.strptime(args.date_end, '%Y-%m-%d')
    except ValueError as e:
        print(f"Error parsing dates: {e}")
        print("Please use YYYY-MM-DD format for dates")
        return

    # Call the function
    concat_orb_clk(
        date_srt=date_srt,
        date_end=date_end,
        nprocs=args.nprocs,
        prod=args.prod,
        verbose=args.verbose
    )

if __name__ == "__main__":
    main()